""" Configuration

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-01-30
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

import os


class Config(object):
    """ Configuration

    Attributes:
        pull_docker_image (:obj:`bool`): whether to pull the Docker image for the simulator (default: :obj:`True`)
        docker_hub_username (:obj:`str`): username for Docker Hub
        docker_hub_token (:obj:`str`): Token for Docker Hub
        biosimulators_auth_endpoint (:obj:`str`): Authentification endpoint for the BioSimulators API
        biosimulators_audience (:obj:`str`): Open API audience for the BioSimulators API
        biosimulators_api_client_id (:obj:`str`): Client id of the BioSimulators API
        biosimulators_api_client_secret (:obj:`str`): Client secret of the BioSimulators API
        biosimulators_api_endpoint (:obj:`str`): Base URL for the BioSimulators API
        biosimulators_curator_gh_ids (:obj:`list` of :obj:`str`): GitHub user ids of the BioSimulators curators
        biosimulators_default_specifications_version (:obj:`str`): Default version of the BioSimulators simulation specifications
        biosimulators_default_image_version (:obj:`str`): Default version of the BioSimulators Docker image format
        biosimulators_docker_registry_url (:obj:`str`): URL of the Docker registry for simulators
        biosimulators_docker_registry_username (:obj:`str`): username for Docker registry for simulators
        biosimulators_docker_registry_token (:obj:`str`): token for Docker registry for simulators
        biosimulators_docker_image_url_pattern (:obj:`str`): URL pattern for Docker images for simulators
        runbiosimulations_auth_endpoint (:obj:`str`): Authentification endpoint for the runBioSimulations API
        runbiosimulations_audience (:obj:`str`): Open API audience for the runBioSimulations API
        runbiosimulations_api_client_id (:obj:`str`): Client id of the runBioSimulations API
        runbiosimulations_api_client_secret (:obj:`str`): Client secret of the runBioSimulations API
        runbiosimulations_api_endpoint (:obj:`str`): Base URL for the runBioSimulations API
        test_case_timeout (:obj:`int`): time out for test cases in seconds
        user_to_exec_in_simulator_containers (:obj:`str` or :obj:`None`): user id or name to execute calls inside simulator containers

            * Use ``_CURRENT_USER_`` to indicate that the Docker container should execute commands as the current user (``os.getuid()``)
            * Use the format ``<name|uid>[:<group|gid>]`` to indicate any other user/group that the Docker container should use to
              execute commands
    """

    def __init__(self,
                 pull_docker_image=None, docker_hub_username=None, docker_hub_token=None,
                 biosimulators_auth_endpoint=None, biosimulators_audience=None,
                 biosimulators_api_client_id=None, biosimulators_api_client_secret=None,
                 biosimulators_api_endpoint=None,
                 biosimulators_curator_gh_ids=None, biosimulators_default_specifications_version=None,
                 biosimulators_default_image_version=None, biosimulators_docker_registry_url=None,
                 biosimulators_docker_registry_username=None, biosimulators_docker_registry_token=None,
                 biosimulators_docker_image_url_pattern=None,
                 runbiosimulations_auth_endpoint=None, runbiosimulations_audience=None,
                 runbiosimulations_api_client_id=None, runbiosimulations_api_client_secret=None,
                 runbiosimulations_api_endpoint=None,
                 test_case_timeout=None,
                 user_to_exec_in_simulator_containers=None):
        """
        Args:
            pull_docker_image (:obj:`bool`, optional): whether to pull the Docker image for the simulator (default: :obj:`True`)
            docker_hub_username (:obj:`str`, optional): username for Docker Hub
            docker_hub_token (:obj:`str`, optional): Token for Docker Hub
            biosimulators_auth_endpoint (:obj:`str`, optional): Authentification endpoint for the BioSimulators API
            biosimulators_audience (:obj:`str`, optional): Open API audience for the BioSimulators API
            biosimulators_api_client_id (:obj:`str`, optional): Client id of the BioSimulators API
            biosimulators_api_client_secret (:obj:`str`, optional): Client secret of the BioSimulators API
            biosimulators_api_endpoint (:obj:`str`, optional): Base URL for the BioSimulators API
            biosimulators_curator_gh_ids (:obj:`list` of :obj:`str`, optional): GitHub user ids of the BioSimulators curators
            biosimulators_default_specifications_version (:obj:`str`, optional): Default version of the BioSimulators simulation specifications
            biosimulators_default_image_version (:obj:`str`, optional): Default version of the BioSimulators Docker image format
            biosimulators_docker_registry_url (:obj:`str`, optional): URL of the Docker registry for simulators
            biosimulators_docker_registry_username (:obj:`str`, optional): username for Docker registry for simulators
            biosimulators_docker_registry_token (:obj:`str`, optional): token for Docker registry for simulators
            biosimulators_docker_image_url_pattern (:obj:`str`, optional): URL pattern for Docker images for simulators
            runbiosimulations_auth_endpoint (:obj:`str`, optional): Authentification endpoint for the runBioSimulations API
            runbiosimulations_audience (:obj:`str`, optional): Open API audience for the runBioSimulations API
            runbiosimulations_api_client_id (:obj:`str`, optional): Client id of the runBioSimulations API
            runbiosimulations_api_client_secret (:obj:`str`, optional): Client secret of the runBioSimulations API
            runbiosimulations_api_endpoint (:obj:`str`, optional): Base URL for the runBioSimulations API
            test_case_timeout (:obj:`int`, optional): time out for test cases in seconds
            user_to_exec_in_simulator_containers (:obj:`str`, optional): user id or name to execute calls inside simulator containers

                * Use ``_CURRENT_USER_`` to indicate that the Docker container should execute commands as the current user (``os.getuid()``)
                * Use the format ``<name|uid>[:<group|gid>]`` to indicate any other user/group that the Docker container should use to
                  execute commands
        """
        # Docker registry
        if pull_docker_image is None:
            self.pull_docker_image = os.getenv('PULL_DOCKER_IMAGE', '1').lower() in ['1', 'true']
        else:
            self.pull_docker_image = pull_docker_image

        # Docker Hub
        if docker_hub_username is None:
            self.docker_hub_username = os.getenv('DOCKER_HUB_USERNAME')
        else:
            self.docker_hub_username = docker_hub_username

        if docker_hub_token is None:
            self.docker_hub_token = os.getenv('DOCKER_HUB_TOKEN')
        else:
            self.docker_hub_token = docker_hub_token

        # BioSimulators
        if biosimulators_auth_endpoint is None:
            self.biosimulators_auth_endpoint = os.getenv('BIOSIMULATORS_AUTH_ENDPOINT', 'https://auth.biosimulations.org/oauth/token')
        else:
            self.biosimulators_auth_endpoint = biosimulators_auth_endpoint

        if biosimulators_audience is None:
            self.biosimulators_audience = os.getenv('BIOSIMULATORS_AUDIENCE', 'api.biosimulators.org')
        else:
            self.biosimulators_audience = biosimulators_audience

        if biosimulators_api_client_id is None:
            self.biosimulators_api_client_id = os.getenv('BIOSIMULATORS_API_CLIENT_ID')
        else:
            self.biosimulators_api_client_id = biosimulators_api_client_id

        if biosimulators_api_client_secret is None:
            self.biosimulators_api_client_secret = os.getenv('BIOSIMULATORS_API_CLIENT_SECRET')
        else:
            self.biosimulators_api_client_secret = biosimulators_api_client_secret

        if biosimulators_api_endpoint is None:
            self.biosimulators_api_endpoint = os.getenv('BIOSIMULATORS_API_ENDPOINT', 'https://api.biosimulators.org/')
        else:
            self.biosimulators_api_endpoint = biosimulators_api_endpoint

        if biosimulators_curator_gh_ids is None:
            ids = os.getenv('BIOSIMULATORS_CURATOR_GH_IDS', 'jonrkarr').strip()
            if ids:
                self.biosimulators_curator_gh_ids = [id.strip() for id in ids.split(',')]
            else:
                self.biosimulators_curator_gh_ids = []
        else:
            self.biosimulators_curator_gh_ids = biosimulators_curator_gh_ids

        if biosimulators_default_specifications_version is None:
            self.biosimulators_default_specifications_version = os.getenv('BIOSIMULATORS_DEFAULT_SPECIFICATIONS_VERSION', '1.0.0')
        else:
            self.biosimulators_default_specifications_version = biosimulators_default_specifications_version

        if biosimulators_default_image_version is None:
            self.biosimulators_default_image_version = os.getenv('BIOSIMULATORS_DEFAULT_IMAGE_VERSION', '1.0.0')
        else:
            self.biosimulators_default_image_version = biosimulators_default_image_version

        if biosimulators_docker_registry_url is None:
            self.biosimulators_docker_registry_url = os.getenv(
                'BIOSIMULATORS_DOCKER_REGISTRY_URL', os.getenv('DOCKER_REGISTRY_URL', 'ghcr.io'))
        else:
            self.biosimulators_docker_registry_url = biosimulators_docker_registry_url

        if biosimulators_docker_registry_username is None:
            self.biosimulators_docker_registry_username = os.getenv(
                'BIOSIMULATORS_DOCKER_REGISTRY_USERNAME', os.getenv('DOCKER_REGISTRY_USERNAME'))
        else:
            self.biosimulators_docker_registry_username = biosimulators_docker_registry_username

        if biosimulators_docker_registry_token is None:
            self.biosimulators_docker_registry_token = os.getenv(
                'BIOSIMULATORS_DOCKER_REGISTRY_TOKEN', os.getenv('DOCKER_REGISTRY_TOKEN'))
        else:
            self.biosimulators_docker_registry_token = biosimulators_docker_registry_token

        if biosimulators_docker_image_url_pattern is None:
            self.biosimulators_docker_image_url_pattern = os.getenv(
                'BIOSIMULATORS_DOCKER_REGISTRY_IMAGE_URL_PATTERN', 'ghcr.io/biosimulators/{}:{}')
        else:
            self.biosimulators_docker_image_url_pattern = biosimulators_docker_image_url_pattern

        # runBioSimulations
        if runbiosimulations_auth_endpoint is None:
            self.runbiosimulations_auth_endpoint = os.getenv(
                'RUNBIOSIMULATIONS_AUTH_ENDPOINT', 'https://auth.biosimulations.org/oauth/token')
        else:
            self.runbiosimulations_auth_endpoint = runbiosimulations_auth_endpoint

        if runbiosimulations_audience is None:
            self.runbiosimulations_audience = os.getenv('RUNBIOSIMULATIONS_AUDIENCE', 'dispatch.biosimulations.org')
        else:
            self.runbiosimulations_audience = runbiosimulations_audience

        if runbiosimulations_api_client_id is None:
            self.runbiosimulations_api_client_id = os.getenv('RUNBIOSIMULATIONS_API_CLIENT_ID')
        else:
            self.runbiosimulations_api_client_id = runbiosimulations_api_client_id

        if runbiosimulations_api_client_secret is None:
            self.runbiosimulations_api_client_secret = os.getenv('RUNBIOSIMULATIONS_API_CLIENT_SECRET')
        else:
            self.runbiosimulations_api_client_secret = runbiosimulations_api_client_secret

        if runbiosimulations_api_endpoint is None:
            self.runbiosimulations_api_endpoint = os.getenv('RUNBIOSIMULATIONS_API_ENDPOINT', 'https://run.api.biosimulations.org/')
        else:
            self.runbiosimulations_api_endpoint = runbiosimulations_api_endpoint

        if test_case_timeout is None:
            self.test_case_timeout = int(os.getenv('TEST_CASE_TIMEOUT', '300'))  # 300 seconds
        else:
            self.test_case_timeout = test_case_timeout

        if user_to_exec_in_simulator_containers is None:
            self.user_to_exec_in_simulator_containers = os.getenv('USER_TO_EXEC_IN_SIMULATOR_CONTAINERS', '_CURRENT_USER_') or None
        else:
            self.user_to_exec_in_simulator_containers = user_to_exec_in_simulator_containers



""" Utilities for validate containerized simulators

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-21
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from .config import Config
from .data_model import TestCase, OutputMedium
from .exceptions import SkippedTestCaseException, TimeoutException
from .results.data_model import TestCaseResult, TestCaseResultType
from .test_case import cli
from .test_case import combine_archive
from .test_case import docker_image
from .test_case import log
from .test_case import published_project
from .test_case import results_report
from .test_case import sedml
from .warnings import TestCaseWarning, IgnoredTestCaseWarning
from biosimulators_utils.config import Colors
from biosimulators_utils.log.utils import StandardOutputErrorCapturer
import biosimulators_utils.simulator.io
import collections
import contextlib
import datetime
import inspect
import os
import signal
import sys
import termcolor
import warnings

__all__ = ['SimulatorValidator']


class SimulatorValidator(object):
    """ Validate that a Docker image for a simulator implements the BioSimulations simulator interface by
    checking that the image produces the correct outputs for one of more test cases (e.g., COMBINE archive)

    Attributes:
        specifications (:obj:`str` or :obj:`dict`): path or URL to the specifications of the simulator, or the specifications of the simulator
        cases (:obj:`collections.OrderedDict` of :obj:`types.ModuleType` to :obj:`TestCase`): groups of test cases
        verbose (:obj:`bool`): if :obj:`True`, display stdout/stderr from executing cases in real time
        synthetic_archives_dir (:obj:`str`): Directory to save the synthetic COMBINE/OMEX archives generated by the test cases
        output_medium (:obj:`OutputMedium`): environment where outputs will be sent
        log_std_out_err (:obj:`bool`): whether to log the standard output and error generated by each test case
        dry_run (:obj:`bool`): if :obj:`True`, do not use the simulator to execute COMBINE/OMEX archives.
        cli (:obj:`str`): command-line interface to use to execute the tests involving the simulation of COMBINE/OMEX
            archives rather than a Docker image
    """

    def __init__(self, specifications, case_ids=None, verbose=False, synthetic_archives_dir=None, output_medium=OutputMedium.console,
                 log_std_out_err=True, dry_run=False, cli=None, validate_specs=True):
        """
        Args:
            specifications (:obj:`str` or :obj:`dict`): path or URL to the specifications of the simulator, or the specifications of the simulator
            case_ids (:obj:`list` of :obj:`str`, optional): List of ids of test cases to verify. If :obj:`ids`
                is none, all test cases are verified.
            verbose (:obj:`bool`, optional): if :obj:`True`, display stdout/stderr from executing cases in real time
            synthetic_archives_dir (:obj:`str`, optional): Directory to save the synthetic COMBINE/OMEX archives generated by the test cases
            output_medium (:obj:`OutputMedium`, optional): environment where outputs will be sent
            log_std_out_err (:obj:`bool`, optional): whether to log the standard output and error generated by each test case
            dry_run (:obj:`bool`, optional): if :obj:`True`, do not use the simulator to execute COMBINE/OMEX archives.
            cli (:obj:`str`, optional): command-line interface to use to execute the tests involving the simulation of COMBINE/OMEX
                archives rather than a Docker image
            validate_specs (:obj:`bool`, optional): whether to validate specifications
        """
        # if necessary, get and validate specifications of simulator
        if isinstance(specifications, str):
            specifications = biosimulators_utils.simulator.io.read_simulator_specs(specifications, validate=validate_specs)

        self.specifications = specifications
        self.verbose = verbose
        if synthetic_archives_dir and not os.path.isdir(synthetic_archives_dir):
            os.makedirs(synthetic_archives_dir)
        self.synthetic_archives_dir = synthetic_archives_dir
        self.output_medium = output_medium
        self.log_std_out_err = log_std_out_err
        self.dry_run = dry_run
        self.cli = cli

        self.cases = self.find_cases(ids=case_ids)

        self.test_case_timeout = Config().test_case_timeout

    def find_cases(self, ids=None):
        """ Find test cases

        Args:
            ids (:obj:`list` of :obj:`str`, optional): List of ids of test cases to verify. If :obj:`ids`
                is none, all test cases are verified.

        Returns:
            :obj:`collections.OrderedDict` of :obj:`types.ModuleType` to :obj:`TestCase`: groups of test cases
        """
        cases = collections.OrderedDict()

        # get cases involving curated published COMBINE/OMEX archives
        all_published_projects_test_cases, compatible_published_projects_test_cases = published_project.find_cases(
            self.specifications, output_medium=self.output_medium)

        # get Docker image cases
        suite_name = docker_image.__name__.replace('biosimulators_test_suite.test_case.', '')
        cases[suite_name] = self.find_cases_in_module(docker_image, compatible_published_projects_test_cases, ids=ids)

        # get command-line interface cases
        suite_name = cli.__name__.replace('biosimulators_test_suite.test_case.', '')
        cases[suite_name] = self.find_cases_in_module(cli, compatible_published_projects_test_cases, ids=ids)

        # get COMBINE archive test cases
        suite_name = combine_archive.__name__.replace('biosimulators_test_suite.test_case.', '')
        cases[suite_name] = self.find_cases_in_module(combine_archive, compatible_published_projects_test_cases, ids=ids)

        # get SED-ML cases
        suite_name = sedml.__name__.replace('biosimulators_test_suite.test_case.', '')
        cases[suite_name] = self.find_cases_in_module(sedml, compatible_published_projects_test_cases, ids=ids)

        # get cases for reports of simulation results
        suite_name = results_report.__name__.replace('biosimulators_test_suite.test_case.', '')
        cases[suite_name] = self.find_cases_in_module(results_report, compatible_published_projects_test_cases, ids=ids)

        # get cases for reporting status of the execution of modeling projects
        suite_name = log.__name__.replace('biosimulators_test_suite.test_case.', '')
        cases[suite_name] = self.find_cases_in_module(log, compatible_published_projects_test_cases, ids=ids)

        # add cases involving published COMBINE/OMEX archives
        suite_name = published_project.__name__.replace('biosimulators_test_suite.test_case.', '')
        cases[suite_name] = []
        for case in all_published_projects_test_cases:
            if ids is None or case.id in ids:
                cases[suite_name].append(case)

        # warn if desired cases weren't found
        if ids is not None:
            found_case_ids = set()
            for suite_cases in cases.values():
                for case in suite_cases:
                    found_case_ids.add(case.id)

            missing_ids = set(ids).difference(found_case_ids)
            if missing_ids:
                warnings.warn('Some test case(s) were not found:\n  {}'.format('\n  '.join(sorted(missing_ids))), IgnoredTestCaseWarning)

        # return discovered cases
        return cases

    def find_cases_in_module(self, module, published_projects_test_cases, ids=None):
        """ Discover test cases in a module

        Args:
            module (:obj:`types.ModuleType`): module
            ids (:obj:`list` of :obj:`str`, optional): List of ids of test cases to verify. If :obj:`ids`
                is none, all test cases are verified.
            published_projects_test_cases (:obj:`list` of :obj:`published_project.SimulatorCanExecutePublishedProject`): test cases involving
                executing curated COMBINE/OMEX archives

        Returns:
            :obj:`list` of :obj:`TestCase`: test cases
        """
        cases = []
        ignored_ids = []
        module_name = module.__name__.replace('biosimulators_test_suite.test_case.', '')
        for child_name in dir(module):
            child = getattr(module, child_name)
            if isinstance(child, type) and issubclass(child, TestCase) and not inspect.isabstract(child):
                id = module_name + '.' + child_name
                if ids is None or id in ids:
                    description = child.__doc__ or None
                    if description:
                        description_lines = (description
                                             .replace('\r', '')
                                             .replace('\n    ', '\n')
                                             .partition('\n\n')[0]
                                             .strip()
                                             .split('\n'))
                        description = ' '.join(line.strip() for line in description_lines) or None
                    if issubclass(child, published_project.SyntheticCombineArchiveTestCase):
                        case = child(id=id, description=description, output_medium=self.output_medium,
                                     published_projects_test_cases=published_projects_test_cases)
                    else:
                        case = child(id=id, description=description, output_medium=self.output_medium)
                    cases.append(case)
                else:
                    ignored_ids.append(id)

        if ignored_ids:
            warnings.warn('Some test case(s) were ignored:\n  {}'.format('\n  '.join(sorted(ignored_ids))), IgnoredTestCaseWarning)

        cases.sort(key=lambda case: case.id)

        return cases

    def run(self):
        """ Validate that a Docker image for a simulator implements the BioSimulations simulator interface by
        checking that the image produces the correct outputs for test cases (e.g., COMBINE archive)

        Returns:
            :obj:`list` :obj:`TestCaseResult`: results of executing test cases
        """
        # print starting message
        n_cases = 0
        for suite_cases in self.cases.values():
            n_cases += len(suite_cases)
        print('Collected {} test cases.'.format(n_cases))

        # get start time
        start = datetime.datetime.now()

        # execute test cases and collect results
        results = []
        for suite_name, suite_cases in self.cases.items():
            print('\nExecuting {} {} tests ... {}'.format(len(suite_cases), suite_name, 'done' if not suite_cases else ''))
            for i_case, case in enumerate(suite_cases):
                print('  {}: {} ... '.format(i_case + 1, case.id), end='')
                sys.stdout.flush()

                result = self.eval_case(case)
                results.append(result)

                print(termcolor.colored(result.type.value, Colors[result.type.value].value), end='')
                print(' (', end='')
                if result.warnings:
                    print(termcolor.colored(str(len(result.warnings)) + ' warnings, ', Colors.warned.value), end='')
                print('{:.1f} s'.format(result.duration), end='')
                print(').')

        # get total duration
        duration = (datetime.datetime.now() - start).total_seconds()

        # print completion message
        print('\n{} tests completed in {:.1f} s'.format(n_cases, duration))

        # return results
        return results

    def eval_case(self, case):
        """ Evaluate a test case for a simulator

        Args:
            case (:obj:`TestCase`): test case

        Returns:
            :obj:`TestCaseResult`: test case result
        """
        start_time = datetime.datetime.now()

        with StandardOutputErrorCapturer(relay=self.verbose, disabled=not self.log_std_out_err) as captured:
            with warnings.catch_warnings(record=True) as caught_warnings:
                warnings.simplefilter("ignore")
                warnings.simplefilter("always", TestCaseWarning)

                try:

                    with time_limit(seconds=self.test_case_timeout):
                        case.eval(self.specifications, synthetic_archives_dir=self.synthetic_archives_dir,
                                  dry_run=self.dry_run, cli=self.cli)
                    type = TestCaseResultType.passed
                    exception = None
                    skip_reason = None

                except SkippedTestCaseException as caught_exception:
                    type = TestCaseResultType.skipped
                    exception = None
                    skip_reason = caught_exception

                except Exception as caught_exception:
                    type = TestCaseResultType.failed
                    exception = caught_exception
                    skip_reason = None

                duration = (datetime.datetime.now() - start_time).total_seconds()

                return TestCaseResult(
                    case=case,
                    type=type,
                    duration=duration,
                    exception=exception,
                    warnings=caught_warnings,
                    skip_reason=skip_reason,
                    log=captured.get_text())

    @staticmethod
    def summarize_results(results):
        """ Get a summary of the results of a set of test cases

        Args:
            results (:obj:`list` :obj:`TestCaseResult`): results of executing test cases

        Returns:
            :obj:`tuple`

                * :obj:`str`: summary of results of test cases
                * :obj:`list` of :obj:`str`: details of failures
                * :obj:`list` of :obj:`str`: details of warnings
                * :obj:`list` of :obj:`str`: details of skips
        """
        passed = []
        failed = []
        skipped = []
        warning_details = []
        failure_details = []
        skipped_details = []
        for result in sorted(results, key=lambda result: result.case.id):
            if result.type == TestCaseResultType.passed:
                result_str = '  * `{}`\n'.format(result.case.id)
                passed.append(result_str)

            elif result.type == TestCaseResultType.failed:
                result_str = '  * `{}`\n'.format(result.case.id)
                failed.append(result_str)

                detail = ''
                detail += '`{}` ({:.1f} s)\n'.format(result.case.id, result.duration)
                detail += '\n'
                if result.case.description:
                    detail += '  {}\n'.format(result.case.description.replace('\n', '\n  '))
                    detail += '\n'

                detail += '  Exception:\n'
                detail += '\n'
                detail += '  ```\n'
                detail += '  {}\n'.format(str(result.exception).replace('\n', '\n  '))
                detail += '  ```\n'
                detail += '\n'

                detail += '  Log:\n'
                detail += '\n'
                detail += '  ```\n'
                detail += '  {}\n'.format(result.log.replace('\n', '\n  ') if result.log else '')
                detail += '  ```'

                failure_details.append(detail)

            elif result.type == TestCaseResultType.skipped:
                result_str = '  * `{}`\n'.format(result.case.id)
                skipped.append(result_str)

                detail = ''
                detail += '`{}` ({:.1f} s)\n'.format(result.case.id, result.duration)
                detail += '\n'
                if result.case.description:
                    detail += '  {}\n'.format(result.case.description.replace('\n', '\n  '))
                    detail += '\n'

                detail += '  Reason for skip:\n'
                detail += '\n'
                detail += '  ```\n'
                detail += '  {}\n'.format(str(result.skip_reason).replace('\n', '\n  '))
                detail += '  ```\n'
                detail += '\n'

                if result.warnings:
                    detail += '  Warnings:\n'
                    for warning in result.warnings:
                        detail += '\n'
                        detail += '  ```\n'
                        detail += '  {}\n'.format(str(warning.message).replace('\n', '\n  '))
                        detail += '  ```\n'
                    detail += '\n'

                detail += '  Log:\n'
                detail += '\n'
                detail += '  ```\n'
                detail += '  {}\n'.format(result.log.replace('\n', '\n  ') if result.log else '')
                detail += '  ```'

                skipped_details.append(detail)

            if result.warnings:
                detail = ''
                detail += '`{}` ({:.1f} s)\n'.format(result.case.id, result.duration)
                detail += '\n'
                if result.case.description:
                    detail += '  {}\n'.format(result.case.description.replace('\n', '\n  '))
                    detail += '\n'

                detail += '  Warnings:\n'
                for warning in result.warnings:
                    detail += '\n'
                    detail += '  ```\n'
                    detail += '  {}\n'.format(str(warning.message).replace('\n', '\n  '))
                    detail += '  ```\n'
                detail += '\n'

                detail += '  Log:\n'
                detail += '\n'
                detail += '  ```\n'
                detail += '  {}\n'.format(result.log.replace('\n', '\n  ') if result.log else '')
                detail += '  ```'

                warning_details.append(detail)

        return (
            '\n'.join([
                '* Executed {} test cases\n'.format(len(results)),
                '* Passed {} test cases{}\n{}'.format(len(passed), ':' if passed else '', ''.join(passed)),
                '* Failed {} test cases{}\n{}'.format(len(failed), ':' if failed else '', ''.join(failed)),
                '* Skipped {} test cases{}\n{}'.format(len(skipped), ':' if skipped else '', ''.join(skipped)),
            ]).strip(),
            failure_details,
            warning_details,
            skipped_details,
        )


@contextlib.contextmanager
def time_limit(seconds):
    """ Context manager for timing out long operations

    Args:
        seconds (:obj:`int`): length in seconds before time out

    Raises:
        :obj:`TimeoutException`: if the operation timed out
    """
    def signal_handler(signum, frame):
        raise TimeoutException("Operation did not complete within {} seconds".format(seconds))
    signal.signal(signal.SIGALRM, signal_handler)
    signal.alarm(int(seconds))
    try:
        yield
    finally:
        signal.alarm(0)



""" Methods for checking support for reports of simulation results

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-29
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..warnings import TestCaseWarning
from .published_project import SingleMasterSedDocumentCombineArchiveTestCase
from biosimulators_utils.combine.data_model import CombineArchive  # noqa: F401
from biosimulators_utils.report.data_model import ReportFormat
from biosimulators_utils.report.io import ReportReader
from biosimulators_utils.sedml.data_model import SedDocument, Report  # noqa: F401
import os
import numpy
import warnings

__all__ = [
    'SimulatorGeneratesReportsOfSimulationResults',
]


class SimulatorGeneratesReportsOfSimulationResults(SingleMasterSedDocumentCombineArchiveTestCase):
    """ Test that when a COMBINE/OMEX archive defines a (single) master file, the simulator only
    executes this file.
    """

    def eval_outputs(self, specifications, synthetic_archive, synthetic_sed_docs, outputs_dir):
        """ Test that the expected outputs were created for the synthetic archive

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            synthetic_archive (:obj:`CombineArchive`): synthetic COMBINE/OMEX archive for testing the simulator
            synthetic_sed_docs (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): map from the location of each SED
                document in the synthetic archive to the document
            outputs_dir (:obj:`str`): directory that contains the outputs produced from the execution of the synthetic archive

        Returns:
            :obj:`bool`: whether there were no warnings about the outputs
        """
        try:
            ReportReader().get_ids(outputs_dir)
        except Exception:
            raise ValueError('Simulator must generate reports of simulation results')

        has_warning = False
        for doc_location, sed_doc in synthetic_sed_docs.items():
            doc_id = os.path.relpath(doc_location, '.')
            for output in sed_doc.outputs:
                if isinstance(output, Report):
                    report_data = ReportReader().run(output, outputs_dir, os.path.join(doc_id, output.id), format=ReportFormat.h5)

                    expected_data_sets = set(data_set.id for data_set in output.data_sets)
                    data_sets = set(report_data.keys())

                    missing_data_sets = expected_data_sets.difference(data_sets)
                    # extra_data_sets = data_sets.difference(expected_data_sets)

                    if missing_data_sets:
                        raise ValueError('Simulator did not produce the following data sets:\n  - {}'.format(
                            '\n  - '.join(sorted(missing_data_sets))))

                    for data_set_data in report_data.values():
                        if numpy.any(numpy.isnan(data_set_data)):
                            warnings.warn('The results produced by the simulator include `NaN`.', TestCaseWarning)
                            has_warning = True

        return not has_warning



""" Common warnings for test cases

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-21
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

__all__ = [
    'TestCaseWarning',
    'IgnoredTestCaseWarning',
    'SimulatorRuntimeErrorWarning',
    'InvalidOutputsWarning',
]


class TestCaseWarning(UserWarning):
    """ Base class for warnings collected from test cases """
    pass


class SimulatorRuntimeErrorWarning(TestCaseWarning, RuntimeWarning):
    """ Warning that the execution of a test case failed """
    pass  # pragma: no co


class IgnoredTestCaseWarning(TestCaseWarning):
    """ Warning that a test case was ignored """
    pass  # pragma: no cover


class InvalidOutputsWarning(TestCaseWarning):
    """ Warning that the outputs of the execution of a COMBINE/OMEX archive were not as expected """
    pass  # pragma: no cover



""" Methods for export test results

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-01-01
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from .data_model import TestCaseResult, TestResultsReport  # noqa: F401
import json

__all__ = ['write_test_results']


def write_test_results(results, filename, gh_issue=None, gh_action_run=None):
    """ Write the results of test cases to a JSON file

    Args:
        results (:obj:`list` of :obj:`TestCaseResult`): results of test cases
        filename (:obj:`str`): path to save results
        gh_issue (:obj:`int`, optional): GitHub issue for which the test suite was executed
        gh_action_run (:obj:`int`, optional): GitHub action run in which the test suite was executed
    """
    report = TestResultsReport(results=results, gh_issue=gh_issue, gh_action_run=gh_action_run)
    with open(filename, 'w') as file:
        json.dump(report.to_dict(), file)



""" Methods for testing that simulators support the features of SED-ML

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-21
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""
from ..exceptions import InvalidOutputsException, SkippedTestCaseException
from ..warnings import InvalidOutputsWarning
from .published_project import SingleMasterSedDocumentCombineArchiveTestCase, UniformTimeCourseTestCase, ExpectedResultOfSyntheticArchive
from biosimulators_utils.combine.data_model import CombineArchive  # noqa: F401
from biosimulators_utils.archive.io import ArchiveReader
from biosimulators_utils.config import get_config
from biosimulators_utils.report.io import ReportReader
from biosimulators_utils.sedml.data_model import (SedDocument, Task, Output, Report, Plot2D, Plot3D,  DataGenerator,  # noqa: F401
                                                  Variable, UniformTimeCourseSimulation,
                                                  DataSet, Curve, Surface, AxisScale,
                                                  Model, ModelAttributeChange, AlgorithmParameterChange,
                                                  AddElementModelChange, RemoveElementModelChange, ReplaceElementModelChange,
                                                  ComputeModelChange, Parameter,
                                                  RepeatedTask, Range, UniformRange, UniformRangeType, VectorRange, FunctionalRange,
                                                  SetValueComputeModelChange, SubTask)
from biosimulators_utils.sedml.utils import get_xml_node_namespace_tag_target
from kisao import Kisao
from kisao.data_model import AlgorithmSubstitutionPolicy
from kisao.utils import get_substitutable_algorithms_for_policy
from lxml import etree
import abc
import copy
import numpy
import numpy.testing
import os
import PyPDF2
import re
import shutil
import tempfile
import warnings

__all__ = [
    'SimulatorSupportsModelsSimulationsTasksDataGeneratorsAndReports',
    'SimulatorSupportsModelAttributeChanges',
    'SimulatorSupportsAlgorithmParameters',
    'SimulatorProducesReportsWithCuratedNumberOfDimensions',
    'SimulatorSupportsMultipleTasksPerSedDocument',
    'SimulatorSupportsMultipleReportsPerSedDocument',
    'SimulatorSupportsUniformTimeCoursesWithNonZeroOutputStartTimes',
    'SimulatorSupportsUniformTimeCoursesWithNonZeroInitialTimes',
    'SimulatorSupportsRepeatedTasksWithLinearUniformRanges',
    'SimulatorSupportsRepeatedTasksWithLogarithmicUniformRanges',
    'SimulatorSupportsRepeatedTasksWithVectorRanges',
    'SimulatorSupportsRepeatedTasksWithFunctionalRanges',
    'SimulatorSupportsRepeatedTasksWithNestedFunctionalRanges',
    'SimulatorSupportsRepeatedTasksWithFunctionalRangeVariables',
    'SimulatorSupportsRepeatedTasksWithMultipleSubTasks',
    'SimulatorSupportsRepeatedTasksWithNestedRepeatedTasks',
    'SimulatorSupportsRepeatedTasksWithSubTasksOfMixedTypes',
    'SimulatorProducesLinear2DPlots',
    'SimulatorProducesLogarithmic2DPlots',
    'SimulatorProducesLinear3DPlots',
    'SimulatorProducesLogarithmic3DPlots',
    'SimulatorProducesMultiplePlots',
    'SimulatorSupportsSubstitutingAlgorithms',
]


class SimulatorSupportsModelsSimulationsTasksDataGeneratorsAndReports(SingleMasterSedDocumentCombineArchiveTestCase):
    """ Test that a simulator supports the core elements of SED: models, simulations, tasks, data generators for
    individual variables, and reports
    """

    def eval_outputs(self, specifications, synthetic_archive, synthetic_sed_docs, outputs_dir):
        """ Test that the expected outputs were created for the synthetic archive

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            synthetic_archive (:obj:`CombineArchive`): synthetic COMBINE/OMEX archive for testing the simulator
            synthetic_sed_docs (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): map from the location of each SED
                document in the synthetic archive to the document
            outputs_dir (:obj:`str`): directory that contains the outputs produced from the execution of the synthetic archive

        Returns:
            :obj:`bool`: whether there were no warnings about the outputs
        """
        has_warnings = False

        # reports
        try:
            report_ids = ReportReader().get_ids(outputs_dir)
        except Exception:
            report_ids = []

        expected_report_ids = set()
        for doc_location, doc in synthetic_sed_docs.items():
            doc_id = os.path.relpath(doc_location, './')
            for output in doc.outputs:
                if isinstance(output, Report):
                    expected_report_ids.add(os.path.join(doc_id, output.id))

        missing_report_ids = expected_report_ids.difference(set(report_ids))
        extra_report_ids = set(report_ids).difference(expected_report_ids)

        if missing_report_ids:
            raise InvalidOutputsException('Simulator did not produce the following reports:\n  - {}'.format(
                '\n  - '.join(sorted('`' + id + '`' for id in missing_report_ids))
            ))

        if extra_report_ids:
            msg = 'Simulator produced extra reports:\n  - {}'.format(
                '\n  - '.join(sorted('`' + id + '`' for id in extra_report_ids)))
            warnings.warn(msg, InvalidOutputsWarning)
            has_warnings = True

        # data sets
        expected_data_set_ids = set()
        data_set_ids = set()
        for doc_location, doc in synthetic_sed_docs.items():
            doc_id = os.path.relpath(doc_location, './')
            for output in doc.outputs:
                if isinstance(output, Report):
                    for data_set in output.data_sets:
                        expected_data_set_ids.add(os.path.join(doc_id, output.id, data_set.id))

                    results = ReportReader().run(output, outputs_dir, os.path.join(doc_id, output.id))
                    data_set_ids.update(set(os.path.join(doc_id, output.id, id) for id in results.keys()))

        missing_data_set_ids = expected_data_set_ids.difference(set(data_set_ids))
        extra_data_set_ids = set(data_set_ids).difference(expected_data_set_ids)

        if missing_data_set_ids:
            raise InvalidOutputsException('Simulator did not produce the following data sets:\n  - {}'.format(
                '\n  - '.join(sorted('`' + id + '`' for id in missing_data_set_ids))
            ))

        if extra_data_set_ids:
            msg = 'Simulator produced extra data sets:\n  - {}'.format(
                '\n  - '.join(sorted('`' + id + '`' for id in extra_data_set_ids)))
            warnings.warn(msg, InvalidOutputsWarning)
            has_warnings = True

        return not has_warnings


class SimulatorCanResolveModelSourcesDefinedByUriFragments(SimulatorSupportsModelsSimulationsTasksDataGeneratorsAndReports):
    """ Test that a simulator can resolve model sources defined by URI fragments (e.g., ``#model1``). """

    REPORT_ERROR_AS_SKIP = True

    def __init__(self, *args, **kwargs):
        super(SimulatorCanResolveModelSourcesDefinedByUriFragments, self).__init__(*args, **kwargs)
        self._model_change_filter = lambda change: False

    def build_synthetic_archives(self, specifications, curated_archive, curated_archive_dir, curated_sed_docs):
        """ Generate a synthetic archive with a copy of each task and each report

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            curated_archive (:obj:`CombineArchive`): curated COMBINE/OMEX archive
            curated_archive_dir (:obj:`str`): directory with the contents of the curated COMBINE/OMEX archive
            curated_sed_docs (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): map from locations to
                SED documents in curated archive

        Returns:
            :obj:`list` of :obj:`ExpectedResultOfSyntheticArchive`
        """
        expected_results_of_synthetic_archives = super(SimulatorCanResolveModelSourcesDefinedByUriFragments, self).build_synthetic_archives(
            specifications, curated_archive, curated_archive_dir, curated_sed_docs)
        curated_sed_docs = expected_results_of_synthetic_archives[0].sed_documents

        # change model source to URI fragment
        doc = list(curated_sed_docs.values())[0]

        model = doc.models[0]
        source_model = Model(id='__source__', source=model.source, language=model.language)
        doc.models.append(source_model)
        model.source = '#' + source_model.id

        return expected_results_of_synthetic_archives


class SimulatorCanResolveModelSourcesDefinedByUriFragmentsAndInheritChanges(
        SimulatorSupportsModelsSimulationsTasksDataGeneratorsAndReports):
    """ Test that a simulator can resolve model sources defined by URI fragments (e.g., ``#model1``) and inherit the
    changes of the model."""

    REPORT_ERROR_AS_SKIP = True

    def __init__(self, *args, **kwargs):
        super(SimulatorCanResolveModelSourcesDefinedByUriFragmentsAndInheritChanges, self).__init__(*args, **kwargs)
        self._model_change_filter = lambda change: isinstance(change, ModelAttributeChange)

    def build_synthetic_archives(self, specifications, curated_archive, curated_archive_dir, curated_sed_docs):
        """ Generate a synthetic archive with a copy of each task and each report

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            curated_archive (:obj:`CombineArchive`): curated COMBINE/OMEX archive
            curated_archive_dir (:obj:`str`): directory with the contents of the curated COMBINE/OMEX archive
            curated_sed_docs (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): map from locations to
                SED documents in curated archive

        Returns:
            :obj:`list` of :obj:`ExpectedResultOfSyntheticArchive`
        """
        expected_results_of_synthetic_archives = super(SimulatorCanResolveModelSourcesDefinedByUriFragmentsAndInheritChanges,
                                                       self).build_synthetic_archives(
            specifications, curated_archive, curated_archive_dir, curated_sed_docs)
        curated_sed_docs = expected_results_of_synthetic_archives[0].sed_documents

        # change model source to URI fragment
        doc = list(curated_sed_docs.values())[0]

        model = doc.models[0]
        source_model = Model(id='__source__', source=model.source, language=model.language, changes=model.changes)
        doc.models.append(source_model)
        model.source = '#' + source_model.id
        model.changes = []

        return expected_results_of_synthetic_archives


class SimulatorSupportsModelAttributeChanges(SimulatorSupportsModelsSimulationsTasksDataGeneratorsAndReports):
    """ Test that a simulator supports changes to the attributes of model elements
    """
    REPORT_ERROR_AS_SKIP = True

    def __init__(self, *args, **kwargs):
        super(SimulatorSupportsModelAttributeChanges, self).__init__(*args, **kwargs)
        self._model_change_filter = lambda change: False

    def build_synthetic_archives(self, specifications, curated_archive, curated_archive_dir, curated_sed_docs):
        """ Generate a synthetic archive with a copy of each task and each report

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            curated_archive (:obj:`CombineArchive`): curated COMBINE/OMEX archive
            curated_archive_dir (:obj:`str`): directory with the contents of the curated COMBINE/OMEX archive
            curated_sed_docs (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): map from locations to
                SED documents in curated archive

        Returns:
            :obj:`list` of :obj:`ExpectedResultOfSyntheticArchive`
        """
        expected_results_of_synthetic_archives = super(SimulatorSupportsModelAttributeChanges, self).build_synthetic_archives(
            specifications, curated_archive, curated_archive_dir, curated_sed_docs)
        curated_archive = expected_results_of_synthetic_archives[0].archive
        curated_sed_docs = expected_results_of_synthetic_archives[0].sed_documents

        # get model
        doc = list(curated_sed_docs.values())[0]
        model = doc.models[0]

        try:
            model_etree = etree.parse(os.path.join(curated_archive_dir, model.source))
        except etree.XMLSyntaxError:
            msg = ('This test is only implemented for XML-based model languages. '
                   'Please contact the BioSimulators Team to discuss implementing tests for additional languages.')
            raise SkippedTestCaseException(msg)

        # add model changes
        model_root = model_etree.getroot()

        sed_docs_1 = copy.deepcopy(curated_sed_docs)
        sed_docs_2 = copy.deepcopy(sed_docs_1)
        doc_1 = list(sed_docs_1.values())[0]
        doc_2 = list(sed_docs_2.values())[0]
        model_1 = doc_1.models[0]
        model_2 = doc_2.models[0]
        nodes = [(model_root, 0, '', {})]
        while nodes:
            node, i_node, parent_target, parent_namespaces = nodes.pop()

            _, _, _, node_target, target_namespaces = get_xml_node_namespace_tag_target(
                node, target_namespaces=parent_namespaces)

            node_target = (
                parent_target
                + '/'
                + node_target
                + '[{}]'.format(i_node + 1)
            )

            for key, value in node.attrib.items():
                if key[0] == '{':
                    ns, _, key = key[1:].rpartition('}')
                    rev_namespaces = {v: k for k, v in node.nsmap.items()}
                    key = rev_namespaces[ns] + ':' + key

                model_1.changes.append(
                    ModelAttributeChange(
                        target=node_target + '/@' + key,
                        target_namespaces=target_namespaces,
                        new_value='x',
                    )
                )

                model_2.changes.append(
                    ModelAttributeChange(
                        target=node_target + '/@' + key,
                        target_namespaces=target_namespaces,
                        new_value='x',
                    )
                )
                model_2.changes.append(
                    ModelAttributeChange(
                        target=node_target + '/@' + key,
                        target_namespaces=target_namespaces,
                        new_value=value,
                    )
                )

            n_children = {}
            for child in node.getchildren():
                _, _, _, child_target, _ = get_xml_node_namespace_tag_target(
                    child, target_namespaces=target_namespaces)

                if child_target not in n_children:
                    n_children[child_target] = 0

                nodes.append((child, n_children[child_target], node_target, target_namespaces))

                n_children[child_target] += 1

        return [
            ExpectedResultOfSyntheticArchive(curated_archive, sed_docs_1, False),
            ExpectedResultOfSyntheticArchive(curated_archive, sed_docs_2, True),
        ]


class SimulatorSupportsComputeModelChanges(SimulatorSupportsModelsSimulationsTasksDataGeneratorsAndReports):
    """ Test that a simulator supports compute model changes
    """
    REPORT_ERROR_AS_SKIP = True

    def __init__(self, *args, **kwargs):
        super(SimulatorSupportsComputeModelChanges, self).__init__(*args, **kwargs)
        self._model_change_filter = lambda change: False

    def build_synthetic_archives(self, specifications, curated_archive, curated_archive_dir, curated_sed_docs):
        """ Generate a synthetic archive with a copy of each task and each report

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            curated_archive (:obj:`CombineArchive`): curated COMBINE/OMEX archive
            curated_archive_dir (:obj:`str`): directory with the contents of the curated COMBINE/OMEX archive
            curated_sed_docs (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): map from locations to
                SED documents in curated archive

        Returns:
            :obj:`list` of :obj:`ExpectedResultOfSyntheticArchive`
        """
        expected_results_of_synthetic_archives = super(SimulatorSupportsComputeModelChanges, self).build_synthetic_archives(
            specifications, curated_archive, curated_archive_dir, curated_sed_docs)
        curated_archive = expected_results_of_synthetic_archives[0].archive
        curated_sed_docs = expected_results_of_synthetic_archives[0].sed_documents

        # get model
        doc = list(curated_sed_docs.values())[0]
        model = doc.models[0]

        try:
            model_etree = etree.parse(os.path.join(curated_archive_dir, model.source))
        except etree.XMLSyntaxError:
            msg = ('This test is only implemented for XML-based model languages. '
                   'Please contact the BioSimulators Team to discuss implementing tests for additional languages.')
            raise SkippedTestCaseException(msg)

        # add model changes
        model_root = model_etree.getroot()

        sed_docs_1 = copy.deepcopy(curated_sed_docs)
        sed_docs_2 = copy.deepcopy(sed_docs_1)
        doc_1 = list(sed_docs_1.values())[0]
        doc_2 = list(sed_docs_2.values())[0]
        model_1 = doc_1.models[0]
        model_2 = doc_2.models[0]
        nodes = [(model_root, 0, '', {})]
        i_change = 0
        while nodes:
            node, i_node, parent_target, parent_namespaces = nodes.pop()

            _, _, _, node_target, target_namespaces = get_xml_node_namespace_tag_target(
                node, target_namespaces=parent_namespaces)

            node_target = (
                parent_target
                + '/'
                + node_target
                + '[{}]'.format(i_node + 1)
            )

            for key, value in node.attrib.items():
                if key[0] == '{':
                    ns, _, key = key[1:].rpartition('}')
                    rev_namespaces = {v: k for k, v in node.nsmap.items()}
                    key = rev_namespaces[ns] + ':' + key

                try:
                    float(value)
                except ValueError:
                    continue

                i_change += 1
                param_id = '__p_{}'.format(i_change)
                var_id = '__var_{}'.format(i_change)
                model_1.changes.append(
                    ComputeModelChange(
                        target=node_target + '/@' + key,
                        target_namespaces=target_namespaces,
                        parameters=[
                            Parameter(id=param_id, value=-1)
                        ],
                        variables=[
                            Variable(
                                id=var_id,
                                target=node_target + '/@' + key,
                                target_namespaces=target_namespaces,
                                model=model_1,
                            )
                        ],
                        math='{} * {}'.format(param_id, var_id),
                    )
                )

                model_2.changes.append(
                    ComputeModelChange(
                        target=node_target + '/@' + key,
                        target_namespaces=target_namespaces,
                        parameters=[
                            Parameter(id=param_id, value=-1)
                        ],
                        variables=[
                            Variable(
                                id=var_id,
                                target=node_target + '/@' + key,
                                target_namespaces=target_namespaces,
                                model=model_2,
                            )
                        ],
                        math='{} * {}'.format(param_id, var_id),
                    )
                )
                model_2.changes.append(
                    ModelAttributeChange(
                        target=node_target + '/@' + key,
                        target_namespaces=target_namespaces,
                        new_value=value,
                    )
                )

            n_children = {}
            for child in node.getchildren():
                _, _, _, child_target, _ = get_xml_node_namespace_tag_target(
                    child, target_namespaces=target_namespaces)

                if child_target not in n_children:
                    n_children[child_target] = 0

                nodes.append((child, n_children[child_target], node_target, target_namespaces))

                n_children[child_target] += 1

        return [
            ExpectedResultOfSyntheticArchive(curated_archive, sed_docs_1, False),
            ExpectedResultOfSyntheticArchive(curated_archive, sed_docs_2, True),
        ]


class SimulatorSupportsAddReplaceRemoveModelElementChanges(SimulatorSupportsModelsSimulationsTasksDataGeneratorsAndReports):
    """ Test that a simulator supports model changes that involve adding, replacing, and removing model elements. """
    REPORT_ERROR_AS_SKIP = True

    def __init__(self, *args, **kwargs):
        super(SimulatorSupportsAddReplaceRemoveModelElementChanges, self).__init__(*args, **kwargs)
        self._model_change_filter = lambda change: False

    def build_synthetic_archives(self, specifications, curated_archive, curated_archive_dir, curated_sed_docs):
        """ Generate a synthetic archive with a copy of each task and each report

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            curated_archive (:obj:`CombineArchive`): curated COMBINE/OMEX archive
            curated_archive_dir (:obj:`str`): directory with the contents of the curated COMBINE/OMEX archive
            curated_sed_docs (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): map from locations to
                SED documents in curated archive

        Returns:
            :obj:`list` of :obj:`ExpectedResultOfSyntheticArchive`
        """
        expected_results_of_synthetic_archives = super(SimulatorSupportsAddReplaceRemoveModelElementChanges, self).build_synthetic_archives(
            specifications, curated_archive, curated_archive_dir, curated_sed_docs)
        curated_archive = expected_results_of_synthetic_archives[0].archive
        curated_sed_docs = expected_results_of_synthetic_archives[0].sed_documents

        # get model
        doc = list(curated_sed_docs.values())[0]
        model = doc.models[0]

        try:
            model_etree = etree.parse(os.path.join(curated_archive_dir, model.source))
        except etree.XMLSyntaxError:
            msg = ('This test is only implemented for XML-based model languages. '
                   'Please contact the BioSimulators Team to discuss implementing tests for additional languages.')
            raise SkippedTestCaseException(msg)

        # add model changes
        model_root = model_etree.getroot()
        _, _, _, root_target, root_namespaces = get_xml_node_namespace_tag_target(model_root)

        sed_docs_1 = copy.deepcopy(curated_sed_docs)
        doc_1 = list(sed_docs_1.values())[0]
        model_1 = doc_1.models[0]
        for child in model_root.getchildren():
            _, _, _, child_target, target_namespaces = get_xml_node_namespace_tag_target(
                child, target_namespaces=root_namespaces)
            model_1.changes.append(
                RemoveElementModelChange(
                    target='/' + root_target + '/' + child_target,
                    target_namespaces=target_namespaces,
                )
            )

        sed_docs_2 = copy.deepcopy(sed_docs_1)
        doc_2 = list(sed_docs_2.values())[0]
        model_2 = doc_2.models[0]
        for child in model_root.getchildren():
            model_2.changes.append(
                AddElementModelChange(
                    target='/' + root_target,
                    target_namespaces=root_namespaces,
                    new_elements=etree.tostring(child).decode(),
                )
            )

        sed_docs_3 = copy.deepcopy(curated_sed_docs)
        doc_3 = list(sed_docs_3.values())[0]
        model_3 = doc_3.models[0]
        for child in model_root.getchildren():
            child_uri, child_prefix, _, child_target, target_namespaces = get_xml_node_namespace_tag_target(
                child, target_namespaces=root_namespaces)
            model_3.changes.append(
                ReplaceElementModelChange(
                    target='/' + root_target + '/' + child_target,
                    target_namespaces=target_namespaces,
                    new_elements='<biosimulatorsTestSuite:node xmlns:biosimulatorsTestSuite="https://biosimulatos.org" />',
                )
            )

        return [
            ExpectedResultOfSyntheticArchive(curated_archive, sed_docs_1, False),
            ExpectedResultOfSyntheticArchive(curated_archive, sed_docs_2, True),
            ExpectedResultOfSyntheticArchive(curated_archive, sed_docs_3, False),
        ]


class SimulatorSupportsAlgorithmParameters(SimulatorSupportsModelsSimulationsTasksDataGeneratorsAndReports):
    """ Test that a simulator supports setting the values of parameters of algorithms """

    def is_curated_sed_algorithm_suitable_for_building_synthetic_archive(self, specifications, algorithm):
        """ Determine if a SED algorithm is suitable for testing

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            algorithm (:obj:`Algorithm`): SED algorithm in curated archive

        Returns:
            :obj:`bool`: whether the algorithm is suitable for testing
        """
        for alg_specs in specifications['algorithms']:
            if alg_specs['kisaoId']['id'] == algorithm.kisao_id:
                for param_spec in alg_specs['parameters']:
                    if param_spec['value'] is not None:
                        return True

        return False

    def build_synthetic_archives(self, specifications, curated_archive, curated_archive_dir, curated_sed_docs):
        """ Generate a synthetic archive with a copy of each task and each report

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            curated_archive (:obj:`CombineArchive`): curated COMBINE/OMEX archive
            curated_archive_dir (:obj:`str`): directory with the contents of the curated COMBINE/OMEX archive
            curated_sed_docs (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): map from locations to
                SED documents in curated archive

        Returns:
            :obj:`list` of :obj:`ExpectedResultOfSyntheticArchive`
        """
        expected_results_of_synthetic_archives = super(SimulatorSupportsAlgorithmParameters, self).build_synthetic_archives(
            specifications, curated_archive, curated_archive_dir, curated_sed_docs)
        curated_sed_docs = expected_results_of_synthetic_archives[0].sed_documents

        # set algorithm parameters
        doc = list(curated_sed_docs.values())[0]

        algorithm = doc.simulations[0].algorithm

        for alg_specs in specifications['algorithms']:
            if alg_specs['kisaoId']['id'] == algorithm.kisao_id:
                if alg_specs['parameters']:
                    break

        algorithm.changes = []
        for param_spec in alg_specs['parameters']:
            if param_spec['value'] is not None:
                algorithm.changes.append(
                    AlgorithmParameterChange(
                        kisao_id=param_spec['kisaoId']['id'],
                        new_value=param_spec['value'],
                    )
                )

        return expected_results_of_synthetic_archives


class SimulatorProducesReportsWithCuratedNumberOfDimensions(SimulatorSupportsModelsSimulationsTasksDataGeneratorsAndReports):
    """ Test that that the curated number of output dimensions matches the actual number of output dimensions
    """

    def is_curated_sed_algorithm_suitable_for_building_synthetic_archive(self, specifications, algorithm):
        """ Determine if a SED algorithm is suitable for testing

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            algorithm (:obj:`Algorithm`): SED algorithm in curated archive

        Returns:
            :obj:`bool`: whether the algorithm is suitable for testing
        """
        if not (super(SimulatorProducesReportsWithCuratedNumberOfDimensions, self).
                is_curated_sed_algorithm_suitable_for_building_synthetic_archive(specifications, algorithm)):
            return False

        for alg_specs in specifications['algorithms']:
            if alg_specs['kisaoId']['id'] == algorithm.kisao_id:
                if alg_specs.get('dependentDimensions', None) is not None:
                    return True

        return False

    def eval_outputs(self, specifications, synthetic_archive, synthetic_sed_docs, outputs_dir):
        """ Test that the expected outputs were created for the synthetic archive

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            synthetic_archive (:obj:`CombineArchive`): synthetic COMBINE/OMEX archive for testing the simulator
            synthetic_sed_docs (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): map from the location of each SED
                document in the synthetic archive to the document
            outputs_dir (:obj:`str`): directory that contains the outputs produced from the execution of the synthetic archive

        Returns:
            :obj:`bool`: whether there were no warnings about the outputs
        """
        doc = list(synthetic_sed_docs.values())[0]
        doc_location = list(synthetic_sed_docs.keys())[0]
        doc_id = os.path.relpath(doc_location, './')

        report = doc.outputs[0]
        data = ReportReader().run(report, outputs_dir, os.path.join(doc_id, report.id))

        for alg_specs in specifications['algorithms']:
            if alg_specs['kisaoId']['id'] == doc.simulations[0].algorithm.kisao_id:
                break

        expected_dims = alg_specs['dependentDimensions']

        data_set_data = data[report.data_sets[0].id]
        if numpy.squeeze(data_set_data).ndim != len(expected_dims):
            msg = ('The specifications for the number of dimensions of each data set of algorithm `{}` differs '
                   'from the actual number of dimensions, {} != {}.').format(
                doc.simulations[0].algorithm.kisao_id, numpy.squeeze(data_set_data).ndim, len(expected_dims))
            warnings.warn(msg, InvalidOutputsWarning)
            return False
        else:
            return True


class SimulatorSupportsMultipleTasksPerSedDocument(SingleMasterSedDocumentCombineArchiveTestCase):
    """ Test that a simulator supports multiple tasks per SED document

    Attributes:
        _expected_reports (:obj:`list` of :obj:`tuple` of :obj:`str`): list of pairs of
            original reports and their expected duplicates
    """

    def build_synthetic_archives(self, specifications, curated_archive, curated_archive_dir, curated_sed_docs):
        """ Generate a synthetic archive with a copy of each task and each report

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            curated_archive (:obj:`CombineArchive`): curated COMBINE/OMEX archive
            curated_archive_dir (:obj:`str`): directory with the contents of the curated COMBINE/OMEX archive
            curated_sed_docs (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): map from locations to
                SED documents in curated archive

        Returns:
            :obj:`list` of :obj:`ExpectedResultOfSyntheticArchive`
        """
        expected_results_of_synthetic_archives = super(SimulatorSupportsMultipleTasksPerSedDocument, self).build_synthetic_archives(
            specifications, curated_archive, curated_archive_dir, curated_sed_docs)
        curated_sed_docs = expected_results_of_synthetic_archives[0].sed_documents

        # get a suitable SED document to modify
        location = list(curated_sed_docs.keys())[0]
        doc_id = os.path.relpath(location, './')
        sed_doc = curated_sed_docs[location]

        # duplicate tasks and reports
        copy_tasks = {}
        for task in list(sed_doc.tasks):
            copy_task = copy.copy(task)
            copy_task.id += '__test_suite_copy'
            sed_doc.tasks.append(copy_task)
            copy_tasks[task.id] = copy_task

        self._expected_reports = []
        copy_data_gens = {}
        for output in list(sed_doc.outputs):
            if isinstance(output, Report):
                copy_output = Report(id=output.id + '__test_suite_copy')
                sed_doc.outputs.append(copy_output)

                self._expected_reports.append((
                    os.path.join(doc_id, output.id),
                    os.path.join(doc_id, copy_output.id)))

                for data_set in output.data_sets:
                    copy_data_set = DataSet(id=data_set.id + '__test_suite_copy', label=data_set.label)
                    copy_output.data_sets.append(copy_data_set)

                    data_gen = data_set.data_generator
                    copy_data_gen_id = data_gen.id + '__test_suite_copy'
                    copy_data_gen = copy_data_gens.get(copy_data_gen_id, None)
                    if not copy_data_gen:
                        copy_data_gen = DataGenerator(id=copy_data_gen_id, parameters=data_gen.parameters, math=data_gen.math)
                        sed_doc.data_generators.append(copy_data_gen)

                        for var in data_set.data_generator.variables:
                            copy_var = Variable(
                                id=var.id + '__test_suite_copy',
                                target=var.target,
                                target_namespaces=var.target_namespaces,
                                symbol=var.symbol,
                                model=var.model,
                            )
                            copy_var.task = copy_tasks[var.task.id]
                            copy_data_gen.variables.append(copy_var)

                            copy_data_gen.math = re.sub(r'((^|\b){}(\b|$))'.format(var.id), copy_var.id, copy_data_gen.math)
                    copy_data_set.data_generator = copy_data_gen

        # return modified SED document
        return expected_results_of_synthetic_archives

    def eval_outputs(self, specifications, synthetic_archive, synthetic_sed_docs, outputs_dir):
        """ Test that the expected outputs were created for the synthetic archive

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            synthetic_archive (:obj:`CombineArchive`): synthetic COMBINE/OMEX archive for testing the simulator
            synthetic_sed_docs (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): map from the location of each SED
                document in the synthetic archive to the document
            outputs_dir (:obj:`str`): directory that contains the outputs produced from the execution of the synthetic archive

        Returns:
            :obj:`bool`: whether there were no warnings about the outputs
        """
        try:
            report_ids = ReportReader().get_ids(outputs_dir)
        except Exception:
            report_ids = []

        missing_reports = []
        for report_loc, copy_report_loc in self._expected_reports:
            if report_loc not in report_ids:
                missing_reports.append(report_loc)
            if copy_report_loc not in report_ids:
                missing_reports.append(copy_report_loc)

        if missing_reports:
            raise ValueError('Reports for duplicate tasks were not generated:\n  {}'.format(
                '\n  '.join(sorted(missing_reports))))

        return True


class SimulatorSupportsMultipleReportsPerSedDocument(SingleMasterSedDocumentCombineArchiveTestCase):
    """ Test that a simulator supports multiple reports per SED document """

    def build_synthetic_archives(self, specifications, curated_archive, curated_archive_dir, curated_sed_docs):
        """ Generate a synthetic archive with a copy of each task and each report

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            curated_archive (:obj:`CombineArchive`): curated COMBINE/OMEX archive
            curated_archive_dir (:obj:`str`): directory with the contents of the curated COMBINE/OMEX archive
            curated_sed_docs (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): map from locations to
                SED documents in curated archive

        Returns:
            :obj:`list` of :obj:`ExpectedResultOfSyntheticArchive`
        """
        expected_results_of_synthetic_archives = super(SimulatorSupportsMultipleReportsPerSedDocument, self).build_synthetic_archives(
            specifications, curated_archive, curated_archive_dir, curated_sed_docs)
        curated_sed_docs = expected_results_of_synthetic_archives[0].sed_documents

        # get a suitable SED document to modify
        sed_doc = list(curated_sed_docs.values())[0]

        # divide data sets between two reports
        original_data_sets = sed_doc.outputs[0].data_sets
        sed_doc.outputs = [
            Report(id='report_1'),
            Report(id='report_2'),
        ]

        for i_dataset, data_set in enumerate(original_data_sets):
            sed_doc.outputs[i_dataset % 2].data_sets.append(data_set)

        # to ensure that each report has at least one data set, including when there's only 1 data set total
        sed_doc.outputs[1].data_sets.append(DataSet(
            id=original_data_sets[0].id + '__copy__',
            label=original_data_sets[0].label,
            data_generator=original_data_sets[0].data_generator,
        ))

        # return modified SED document
        return expected_results_of_synthetic_archives

    def eval_outputs(self, specifications, synthetic_archive, synthetic_sed_docs, outputs_dir):
        """ Test that the expected outputs were created for the synthetic archive

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            synthetic_archive (:obj:`CombineArchive`): synthetic COMBINE/OMEX archive for testing the simulator
            synthetic_sed_docs (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): map from the location of each SED
                document in the synthetic archive to the document
            outputs_dir (:obj:`str`): directory that contains the outputs produced from the execution of the synthetic archive

        Returns:
            :obj:`bool`: whether there were no warnings about the outputs
        """
        has_warnings = False

        try:
            report_ids = ReportReader().get_ids(outputs_dir)
        except Exception:
            report_ids = []

        doc_location = os.path.relpath(list(synthetic_sed_docs.keys())[0], './')
        expected_report_ids = set([os.path.join(doc_location, 'report_1'), os.path.join(doc_location, 'report_2')])

        missing_report_ids = expected_report_ids.difference(set(report_ids))
        extra_report_ids = set(report_ids).difference(expected_report_ids)

        if missing_report_ids:
            raise InvalidOutputsException('Simulator did not produce the following reports:\n  - {}'.format(
                '\n  - '.join(sorted('`' + id + '`' for id in missing_report_ids))
            ))

        if extra_report_ids:
            msg = 'Simulator produced extra reports:\n  - {}'.format(
                '\n  - '.join(sorted('`' + id + '`' for id in extra_report_ids)))
            warnings.warn(msg, InvalidOutputsWarning)
            has_warnings = True

        return not has_warnings


class SimulatorSupportsUniformTimeCoursesWithNonZeroOutputStartTimes(UniformTimeCourseTestCase):
    """ Test that a simulator supports time courses with non-zero output start times """

    def modify_simulation(self, simulation):
        """ Modify a simulation

        Args:
            simulation (:obj:`UniformTimeCourseSimulation`): simulation
        """
        simulation.output_start_time = simulation.output_end_time / 2
        simulation.number_of_points = int(simulation.number_of_points / 2)


class SimulatorSupportsUniformTimeCoursesWithNonZeroInitialTimes(UniformTimeCourseTestCase):
    """ Test that a simulator supports multiple time courses with non-zero initial times """

    REPORT_ERROR_AS_SKIP = True

    def modify_simulation(self, simulation):
        """ Modify a simulation

        Args:
            simulation (:obj:`UniformTimeCourseSimulation`): simulation
        """
        simulation.initial_time = simulation.output_end_time / 2
        simulation.output_start_time = simulation.output_end_time / 2
        simulation.number_of_points = int(simulation.number_of_points / 2)


class RepeatedTasksTestCase(SimulatorSupportsModelsSimulationsTasksDataGeneratorsAndReports):
    """ Test that a simulator supports repeated tasks """
    REPORT_ERROR_AS_SKIP = True

    def __init__(self, *args, **kwargs):
        super(RepeatedTasksTestCase, self).__init__(*args, **kwargs)
        self._model_change_filter = lambda change: False

    RANGE_TYPE = VectorRange
    UNIFORM_RANGE_TYPE = UniformRangeType.linear
    RANGE_LENS = (3,)
    FUNCTONAL_RANGE_USES_VARIABLES = False
    NUM_NESTED_RANGES = 0

    def get_ranges(self, i_repeated_task, curated_archive_dir, model, range_len, breaking_range):
        """ Get ranges

        Args:
            i_repeated_task (:obj:`int`): index of repeated task
            curated_archive_dir (:obj:`str`): directory where COMBINE/OMEX archive was unpacked
            model (:obj:`Model`): model
            range_len
            breaking_range (:obj:`bool`): whether the functional range should break the simulation

        Returns:
            :obj:`list` of :obj:`Range`: ranges
        """

        if self.FUNCTONAL_RANGE_USES_VARIABLES:
            try:
                model_etree = etree.parse(os.path.join(curated_archive_dir, model.source))
            except etree.XMLSyntaxError:
                msg = ('This test is only implemented for XML-based model languages. '
                       'Please contact the BioSimulators Team to discuss implementing tests for additional languages.')
                raise SkippedTestCaseException(msg)

            model_root = model_etree.getroot()
            nodes = [(model_root, 0, '', {})]
            variables = []
            while nodes:
                node, i_node, parent_target, parent_namespaces = nodes.pop()

                _, _, _, node_target, target_namespaces = get_xml_node_namespace_tag_target(
                    node, target_namespaces=parent_namespaces)

                node_target = (
                    parent_target
                    + '/'
                    + node_target
                    + '[{}]'.format(i_node + 1)
                )

                for key, value in node.attrib.items():
                    if key[0] == '{':
                        ns, _, key = key[1:].rpartition('}')
                        rev_namespaces = {v: k for k, v in node.nsmap.items()}
                        key = rev_namespaces[ns] + ':' + key

                    try:
                        float(value)
                    except ValueError:
                        continue

                    variables.append(
                        Variable(
                            target=node_target + '/@' + key,
                            target_namespaces=target_namespaces,
                        ),
                    )

                n_children = {}
                for child in node.getchildren():
                    _, _, _, child_target, _ = get_xml_node_namespace_tag_target(
                        child, target_namespaces=target_namespaces)

                    if child_target not in n_children:
                        n_children[child_target] = 0

                    nodes.append((child, n_children[child_target], node_target, target_namespaces))

                    n_children[child_target] += 1

        if self.RANGE_TYPE is UniformRange:
            if self.UNIFORM_RANGE_TYPE == UniformRangeType.linear:
                return [UniformRange(
                    id='__repeated_task_range_' + str(i_repeated_task),
                    start=0.,
                    end=2.,
                    number_of_points=range_len - 1,
                    type=self.UNIFORM_RANGE_TYPE
                )]

            elif self.UNIFORM_RANGE_TYPE == UniformRangeType.log:
                return [UniformRange(
                    id='__repeated_task_range_' + str(i_repeated_task),
                    start=10. ** 0.,
                    end=10. ** 2.,
                    number_of_points=range_len - 1,
                    type=self.UNIFORM_RANGE_TYPE
                )]

        elif self.RANGE_TYPE is VectorRange:
            return [VectorRange(id='__repeated_task_range_' + str(i_repeated_task), values=list(range(range_len)))]

        else:
            ranges = [VectorRange(id='__repeated_task_range_' + str(i_repeated_task), values=list(range(range_len)))]
            for i_range in range(self.NUM_NESTED_RANGES + 1):
                ranges.append(
                    FunctionalRange(
                        id='__repeated_task_range_' + str(i_repeated_task) + '_' + str(i_range + 1),
                        range=ranges[-1],
                        parameters=[
                            Parameter(id='__repeated_task_range_p_' + str(i_repeated_task) + '_' + str(i_range + 1) + '_1', value=0.),
                        ],
                    )
                )
                ranges[-1].math = '{} * {}'.format(ranges[-1].range.id, ranges[-1].parameters[0].id)

                if self.FUNCTONAL_RANGE_USES_VARIABLES:
                    ranges[-1].parameters.append(
                        Parameter(
                            id='__repeated_task_range_p_' + str(i_repeated_task) + '_' + str(i_range + 1) + '_2',
                            value=-1 if breaking_range else 0.,
                        ),
                    )
                    ranges[-1].variables.append(
                        Variable(
                            id='__repeated_task_range_var_' + str(i_repeated_task) + '_' + str(i_range + 1),
                            target=variables[0].target,
                            target_namespaces=variables[0].target_namespaces,
                            model=model,
                        ),
                    )
                    ranges[-1].math += ' + {} * {}'.format(ranges[-1].variables[0].id, ranges[-1].parameters[1].id)
            return ranges

    HAS_CHANGES = False

    def get_changes(self, curated_archive_dir, model, i_repeated_task, range, only_breaking_changes):
        """ Get changes

        Args:
            curated_archive_dir (:obj:`str`): directory where COMBINE/OMEX archive was unpacked
            model (:obj:`Model`): model
            i_repeated_task (:obj:`int`): index of repeated task
            range (:obj:`range`): range
            only_breaking_changes (:obj:`bool`): whether to include breaking or breaking and changes

        Returns:
            :obj:`list` of :obj:`SetValueComputeModelChange`: changes
        """
        if not self.HAS_CHANGES:
            return []

        try:
            model_etree = etree.parse(os.path.join(curated_archive_dir, model.source))
        except etree.XMLSyntaxError:
            msg = ('This test is only implemented for XML-based model languages. '
                   'Please contact the BioSimulators Team to discuss implementing tests for additional languages.')
            raise SkippedTestCaseException(msg)

        model_root = model_etree.getroot()
        nodes = [(model_root, 0, '', {})]
        changes = []
        while nodes:
            node, i_node, parent_target, parent_namespaces = nodes.pop()

            _, _, _, node_target, target_namespaces = get_xml_node_namespace_tag_target(
                node, target_namespaces=parent_namespaces)

            node_target = (
                parent_target
                + '/'
                + node_target
                + '[{}]'.format(i_node + 1)
            )

            for key, value in node.attrib.items():
                if key[0] == '{':
                    ns, _, key = key[1:].rpartition('}')
                    rev_namespaces = {v: k for k, v in node.nsmap.items()}
                    key = rev_namespaces[ns] + ':' + key

                try:
                    original_value = float(value)
                except ValueError:
                    continue

                changes.append(
                    SetValueComputeModelChange(
                        target=node_target + '/@' + key,
                        target_namespaces=target_namespaces,
                        model=model,
                        range=range,
                        parameters=[
                            Parameter(
                                id='p_0_' + str(len(changes)) + '_' + str(i_repeated_task),
                                value=1. if self.FUNCTONAL_RANGE_USES_VARIABLES else 0.,
                            ),
                            Parameter(
                                id='p_1_' + str(len(changes)) + '_' + str(i_repeated_task),
                                value=1. if self.FUNCTONAL_RANGE_USES_VARIABLES else -1,
                            ),
                        ],
                        variables=[
                            Variable(
                                id='var_' + str(len(changes)) + '_' + str(i_repeated_task),
                                model=model,
                                target=node_target + '/@' + key,
                                target_namespaces=target_namespaces,
                            )
                        ],
                        math='{} * {} + {} * {}'.format(
                            'p_0_' + str(len(changes)) + '_' + str(i_repeated_task), range.id,
                            'p_1_' + str(len(changes)) + '_' + str(i_repeated_task), 'var_' +
                            str(len(changes)) + '_' + str(i_repeated_task),
                        )
                    )
                )
                if not only_breaking_changes:
                    changes.append(
                        SetValueComputeModelChange(
                            target=node_target + '/@' + key,
                            target_namespaces=target_namespaces,
                            model=model,
                            range=range,
                            parameters=[
                                Parameter(id='p_2_' + str(len(changes)) + '_' + str(i_repeated_task), value=original_value),
                            ],
                            math='p_2_' + str(len(changes)) + '_' + str(i_repeated_task),
                        )
                    )

            n_children = {}
            for child in node.getchildren():
                _, _, _, child_target, _ = get_xml_node_namespace_tag_target(
                    child, target_namespaces=target_namespaces)

                if child_target not in n_children:
                    n_children[child_target] = 0

                nodes.append((child, n_children[child_target], node_target, target_namespaces))

                n_children[child_target] += 1

        return changes

    NUM_NESTED_REPEATED_TASKS = 0
    MIXED_SUB_TASK_TYPES = False
    NUM_SUB_TASKS = 1

    @abc.abstractmethod
    def is_concrete(self):
        """ Whether the class is abstract

        Returns:
            :obj:`bool`: whether the class is abstract
        """
        pass  # pragma: no cover

    def build_synthetic_archives(self, specifications, curated_archive, curated_archive_dir, curated_sed_docs):
        """ Generate a synthetic archive with a copy of each task and each report

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            curated_archive (:obj:`CombineArchive`): curated COMBINE/OMEX archive
            curated_archive_dir (:obj:`str`): directory with the contents of the curated COMBINE/OMEX archive
            curated_sed_docs (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): map from locations to
                SED documents in curated archive

        Returns:
            :obj:`list` of :obj:`ExpectedResultOfSyntheticArchive`
        """
        expected_results_of_synthetic_archives = super(RepeatedTasksTestCase, self).build_synthetic_archives(
            specifications, curated_archive, curated_archive_dir, curated_sed_docs)

        if self.HAS_CHANGES:
            expected_results_of_synthetic_archive_1 = expected_results_of_synthetic_archives[0]
            expected_results_of_synthetic_archive_2 = copy.deepcopy(expected_results_of_synthetic_archive_1)
            expected_results_of_synthetic_archives.append(expected_results_of_synthetic_archive_2)

            expected_results_of_synthetic_archive_1.is_success_expected = False
            expected_results_of_synthetic_archive_2.is_success_expected = True

            only_breaking_changes = [True, False]
        else:
            if self.MIXED_SUB_TASK_TYPES:
                expected_results_of_synthetic_archive_1 = expected_results_of_synthetic_archives[0]
                expected_results_of_synthetic_archive_2 = copy.deepcopy(expected_results_of_synthetic_archive_1)
                expected_results_of_synthetic_archives.append(expected_results_of_synthetic_archive_2)

                expected_results_of_synthetic_archive_1.is_success_expected = True
                expected_results_of_synthetic_archive_2.is_success_expected = True

                only_breaking_changes = [None, None]
            else:
                only_breaking_changes = [None]

        for i_archive, (expected_results_of_synthetic_archive, only_breaking_change) in enumerate(zip(
                expected_results_of_synthetic_archives, only_breaking_changes)):
            curated_sed_docs = expected_results_of_synthetic_archive.sed_documents

            # get a suitable SED document to modify
            sed_doc = list(curated_sed_docs.values())[0]
            model = sed_doc.models[0]

            # add repeated task
            for i_repeated_task in range(self.NUM_NESTED_REPEATED_TASKS + 1):
                repeated_task = RepeatedTask(id='__repeated_task_' + str(i_repeated_task))
                for i_sub_task in range(self.NUM_SUB_TASKS):
                    if not self.HAS_CHANGES and self.MIXED_SUB_TASK_TYPES and i_sub_task == i_archive:
                        task = sed_doc.tasks[0]
                    else:
                        task = sed_doc.tasks[-1]
                    repeated_task.sub_tasks.append(SubTask(order=self.NUM_SUB_TASKS - i_sub_task, task=task))
                sed_doc.tasks.append(repeated_task)

                repeated_task.ranges = self.get_ranges(i_repeated_task, curated_archive_dir, model,
                                                       self.RANGE_LENS[self.NUM_NESTED_REPEATED_TASKS - i_repeated_task],
                                                       only_breaking_change)
                repeated_task.range = repeated_task.ranges[-1]
                repeated_task.changes = self.get_changes(curated_archive_dir, model,
                                                         i_repeated_task, repeated_task.ranges[-1], only_breaking_change)

            report = sed_doc.outputs[0]
            report.id = 'task_report'
            report_2 = Report(id='__repeated_task_report')
            sed_doc.outputs.append(report_2)
            for i_data_set, data_set in enumerate(report.data_sets):
                data_set_2 = DataSet(id='__repeated_task_data_set_' + str(i_data_set), label=data_set.label)
                report_2.data_sets.append(data_set_2)

                data_gen = data_set.data_generator
                data_gen_2 = DataGenerator(
                    id='__repeated_task_data_generator_' + str(i_data_set),
                    math='__repeated_task_variable_' + str(i_data_set),
                )
                data_set_2.data_generator = data_gen_2
                sed_doc.data_generators.append(data_gen_2)

                variable = data_gen.variables[0]
                variable_2 = Variable(
                    id=data_gen_2.math,
                    symbol=variable.symbol,
                    target=variable.target,
                    target_namespaces=variable.target_namespaces,
                    task=repeated_task)
                data_gen_2.variables.append(variable_2)

        # return modified SED document
        return expected_results_of_synthetic_archives

    def eval_outputs(self, specifications, synthetic_archive, synthetic_sed_docs, outputs_dir):
        """ Test that the expected outputs were created for the synthetic archive

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            synthetic_archive (:obj:`CombineArchive`): synthetic COMBINE/OMEX archive for testing the simulator
            synthetic_sed_docs (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): map from the location of each SED
                document in the synthetic archive to the document
            outputs_dir (:obj:`str`): directory that contains the outputs produced from the execution of the synthetic archive

        Returns:
            :obj:`bool`: whether there were no warnings about the outputs
        """
        doc_location = list(synthetic_sed_docs.keys())[0]
        doc_id = os.path.relpath(doc_location, './')
        doc = synthetic_sed_docs[doc_location]
        report = doc.outputs[0]
        repeated_report = doc.outputs[1]
        results = ReportReader().run(report, outputs_dir, os.path.join(doc_id, report.id))
        repeated_results = ReportReader().run(repeated_report, outputs_dir, os.path.join(doc_id, repeated_report.id))

        for data_set, repeated_data_set in zip(report.data_sets, repeated_report.data_sets):
            results_data_set = results[data_set.id]
            repeated_results_data_set = repeated_results[repeated_data_set.id]

            if results_data_set.ndim == 0:
                results_data_set = results_data_set.reshape((1,))

            if repeated_results_data_set.ndim - results_data_set.ndim != 2 * (self.NUM_NESTED_REPEATED_TASKS + 1):
                raise InvalidOutputsException('Each level of repeated task should contribute two additional dimensions to reports')

            if not self.MIXED_SUB_TASK_TYPES:
                if repeated_results_data_set.shape[0:2 * (self.NUM_NESTED_REPEATED_TASKS + 1):2] != self.RANGE_LENS:
                    raise InvalidOutputsException('The results of the repeated tasks should have a slice for each iteration')

                if not set(repeated_results_data_set.shape[1:2 * (self.NUM_NESTED_REPEATED_TASKS + 1):2]) == set([self.NUM_SUB_TASKS]):
                    raise InvalidOutputsException('The results of the repeated tasks should have a slice for each sub-task')

                if not numpy.any(numpy.isnan(results_data_set)) and numpy.any(numpy.isnan(repeated_results_data_set)):
                    raise InvalidOutputsException('The results of the repeated tasks have unexpected NaNs')
            else:
                shape = results_data_set.shape
                repeated_shape = repeated_results_data_set.shape

                if repeated_shape[0] != self.RANGE_LENS[0]:
                    raise InvalidOutputsException('The results of the repeated tasks should have a slice for each iteration')

                if repeated_shape[1] != self.NUM_SUB_TASKS:
                    raise InvalidOutputsException('The results of the repeated tasks should have a slice for each sub-task')

                if repeated_shape[2] != max(self.RANGE_LENS[1], shape[0]):
                    raise InvalidOutputsException('The results of the repeated tasks should have a slice for each iteration')

                if (
                    (len(shape) == 1 and repeated_shape[3] != self.NUM_SUB_TASKS)
                    or (len(shape) > 1 and repeated_shape[3] != max(self.NUM_SUB_TASKS, shape[1]))
                ):
                    raise InvalidOutputsException('The results of the repeated tasks should have a slice for each sub-task')

                if repeated_shape[2 * (self.NUM_NESTED_REPEATED_TASKS + 1):] != shape:
                    msg = 'The results of the repeated tasks should have a slice for each dimension of output of the basic task'
                    raise InvalidOutputsException(msg)

                sub_tasks = sorted(doc.tasks[-1].sub_tasks, key=lambda sub_task: sub_task.order)
                for i_sub_task, sub_task in enumerate(sub_tasks):
                    if isinstance(sub_task.task, RepeatedTask):
                        slices = tuple([
                            slice(0, repeated_shape[0]),
                            slice(i_sub_task, i_sub_task + 1),
                            slice(0, self.RANGE_LENS[1]),
                            slice(0, self.NUM_SUB_TASKS),
                        ] + [slice(0, dim_len) for dim_len in repeated_shape[4:]])
                        if not numpy.any(numpy.isnan(results_data_set)) and numpy.any(numpy.isnan(repeated_results_data_set[slices])):
                            raise InvalidOutputsException('The results of repeated tasks have unexpected NaNs')

                        slices = tuple([
                            slice(0, repeated_shape[0]),
                            slice(i_sub_task, i_sub_task + 1),
                            slice(self.RANGE_LENS[1], repeated_shape[2]),
                            slice(0, repeated_shape[3]),
                        ] + [slice(0, dim_len) for dim_len in repeated_shape[4:]])
                        if not numpy.all(numpy.isnan(repeated_results_data_set[slices])):
                            raise InvalidOutputsException('The results of repeated tasks have unexpected non-NaNs')

                    else:
                        slices = tuple([
                            slice(0, repeated_shape[0]),
                            slice(i_sub_task, i_sub_task + 1),
                        ] + [
                            slice(0, dim_len) for dim_len in shape
                        ] + [
                            slice(0, 1) for dim_len in repeated_shape[2 + len(shape):]
                        ])
                        if not numpy.any(numpy.isnan(results_data_set)) and numpy.any(numpy.isnan(repeated_results_data_set[slices])):
                            raise InvalidOutputsException('The results of repeated tasks have unexpected NaNs')

                        remaining_dims = repeated_shape[2 + len(shape):]
                        for i_dim, dim_len in enumerate(remaining_dims):
                            slices = [
                                slice(0, repeated_shape[0]),
                                slice(i_sub_task, i_sub_task + 1),
                            ] + [
                                slice(0, dim_len) for dim_len in shape
                            ]
                            slices.extend([slice(0, remaining_dims[ii_dim]) for ii_dim in range(i_dim)])
                            slices.append(slice(1, remaining_dims[i_dim]))
                            slices.extend([slice(0, remaining_dims[ii_dim]) for ii_dim in range(i_dim + 1, len(remaining_dims))])
                            slices = tuple(slices)

                            if not numpy.all(numpy.isnan(repeated_results_data_set[slices])):
                                raise InvalidOutputsException('The results of repeated tasks have unexpected non-NaNs')

        return True


class SimulatorSupportsRepeatedTasksWithLinearUniformRanges(RepeatedTasksTestCase):
    """ Test that a simulator supports repeated tasks over uniform ranges """
    RANGE_TYPE = UniformRange
    UNIFORM_RANGE_TYPE = UniformRangeType.linear
    NUM_SUB_TASKS = 1
    NUM_NESTED_REPEATED_TASKS = 0

    def is_concrete(self): return True


class SimulatorSupportsRepeatedTasksWithLogarithmicUniformRanges(RepeatedTasksTestCase):
    """ Test that a simulator supports repeated tasks over uniform ranges """
    RANGE_TYPE = UniformRange
    UNIFORM_RANGE_TYPE = UniformRangeType.log
    NUM_SUB_TASKS = 1
    NUM_NESTED_REPEATED_TASKS = 0

    def is_concrete(self): return True


class SimulatorSupportsRepeatedTasksWithVectorRanges(RepeatedTasksTestCase):
    """ Test that a simulator supports repeated tasks over vector ranges """
    RANGE_TYPE = VectorRange
    NUM_SUB_TASKS = 1
    NUM_NESTED_REPEATED_TASKS = 0

    def is_concrete(self): return True


class SimulatorSupportsRepeatedTasksWithFunctionalRanges(RepeatedTasksTestCase):
    """ Test that a simulator supports repeated tasks over functional ranges """
    RANGE_TYPE = FunctionalRange
    NUM_NESTED_RANGES = 0
    FUNCTONAL_RANGE_USES_VARIABLES = False
    NUM_SUB_TASKS = 1
    NUM_NESTED_REPEATED_TASKS = 0

    def is_concrete(self): return True


class SimulatorSupportsRepeatedTasksWithNestedFunctionalRanges(RepeatedTasksTestCase):
    """ Test that a simulator supports repeated tasks over nested functional ranges based on model (specification) variables """
    RANGE_TYPE = FunctionalRange
    NUM_NESTED_RANGES = 1
    FUNCTONAL_RANGE_USES_VARIABLES = False
    NUM_SUB_TASKS = 1
    NUM_NESTED_REPEATED_TASKS = 0

    def is_concrete(self): return True


class SimulatorSupportsRepeatedTasksWithFunctionalRangeVariables(RepeatedTasksTestCase):
    """ Test that a simulator supports repeated tasks over nested functional ranges based on model (specification) variables """
    RANGE_TYPE = FunctionalRange
    NUM_NESTED_RANGES = 0
    FUNCTONAL_RANGE_USES_VARIABLES = True
    NUM_SUB_TASKS = 1
    NUM_NESTED_REPEATED_TASKS = 0
    HAS_CHANGES = True

    def is_concrete(self): return True


class SimulatorSupportsRepeatedTasksWithMultipleSubTasks(RepeatedTasksTestCase):
    """ Test that a simulator supports repeated tasks with multiple subtasks """
    RANGE_TYPE = VectorRange
    NUM_SUB_TASKS = 2
    NUM_NESTED_REPEATED_TASKS = 0

    def is_concrete(self): return True


class SimulatorSupportsRepeatedTasksWithChanges(RepeatedTasksTestCase):
    """ Test that a simulator supports repeated tasks with multiple subtasks """
    RANGE_TYPE = VectorRange
    NUM_SUB_TASKS = 1
    NUM_NESTED_REPEATED_TASKS = 0
    HAS_CHANGES = True

    def is_concrete(self): return True


class SimulatorSupportsRepeatedTasksWithNestedRepeatedTasks(RepeatedTasksTestCase):
    """ Test that a simulator supports nested repeated tasks"""
    RANGE_TYPE = VectorRange
    RANGE_LENS = (3, 3)
    NUM_SUB_TASKS = 1
    NUM_NESTED_REPEATED_TASKS = 1
    MIXED_SUB_TASK_TYPES = False

    def is_concrete(self): return True


class SimulatorSupportsRepeatedTasksWithSubTasksOfMixedTypes(RepeatedTasksTestCase):
    """ Test that a simulator supports repeated tasks whose sub-tasks have mixed types.
    Also tests that sub-types executed in order of the values of their ``order`` attributes
    and that reports of the results of repeated tasks handle sub-tasks to produce results of
    different shapes.
    """
    RANGE_TYPE = VectorRange
    RANGE_LENS = (2, 3)
    NUM_SUB_TASKS = 2
    NUM_NESTED_REPEATED_TASKS = 1
    MIXED_SUB_TASK_TYPES = True

    def is_concrete(self): return True


class SimulatorProducesPlotsTestCase(SingleMasterSedDocumentCombineArchiveTestCase):
    """ Test that a simulator produces plots """

    @property
    def _num_plots(self):
        return 1

    @property
    @abc.abstractmethod
    def _axis_scale(self):
        pass  # pragma: no cover

    def build_synthetic_archives(self, specifications, curated_archive, curated_archive_dir, curated_sed_docs):
        """ Generate a synthetic archive with a copy of each task and each report

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            curated_archive (:obj:`CombineArchive`): curated COMBINE/OMEX archive
            curated_archive_dir (:obj:`str`): directory with the contents of the curated COMBINE/OMEX archive
            curated_sed_docs (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): map from locations to
                SED documents in curated archive

        Returns:
            :obj:`list` of :obj:`ExpectedResultOfSyntheticArchive`
        """
        expected_results_of_synthetic_archives = super(SimulatorProducesPlotsTestCase, self).build_synthetic_archives(
            specifications, curated_archive, curated_archive_dir, curated_sed_docs)
        curated_sed_docs = expected_results_of_synthetic_archives[0].sed_documents

        # get a suitable SED document to modify
        doc = list(curated_sed_docs.values())[0]

        # replace report with plot(s)
        doc.outputs = self.build_plots(doc.data_generators)

        # return modified SED document
        return expected_results_of_synthetic_archives

    @abc.abstractmethod
    def build_plots(self, data_generators):
        """ Build plots from the defined data generators

        Args:
            data_generators (:obj:`list` of :obj:`DataGenerator`): data generators

        Returns:
            :obj:`list` of :obj:`Output`: plots
        """
        pass  # pragma: no cover

    def eval_outputs(self, specifications, synthetic_archive, synthetic_sed_docs, outputs_dir):
        """ Test that the expected outputs were created for the synthetic archive

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            synthetic_archive (:obj:`CombineArchive`): synthetic COMBINE/OMEX archive for testing the simulator
            synthetic_sed_docs (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): map from the location of each SED
                document in the synthetic archive to the document
            outputs_dir (:obj:`str`): directory that contains the outputs produced from the execution of the synthetic archive
        """
        plots_path = os.path.join(outputs_dir, get_config().PLOTS_PATH)
        if not os.path.isfile(plots_path):
            raise SkippedTestCaseException('Simulator did not produce plots')

        tempdir = tempfile.mkdtemp()
        try:
            archive = ArchiveReader().run(plots_path, tempdir)
        except Exception:
            shutil.rmtree(tempdir)
            raise InvalidOutputsException('Simulator produced an invalid zip archive of plots')

        for file in archive.files:
            with open(file.local_path, 'rb') as file:
                try:
                    PyPDF2.PdfFileReader(file)
                except Exception:
                    shutil.rmtree(tempdir)
                    raise InvalidOutputsException('Simulator produced an invalid PDF plot')

        doc = list(synthetic_sed_docs.values())[0]
        doc_location = list(synthetic_sed_docs.keys())[0]
        doc_id = os.path.relpath(doc_location, './')

        expected_plot_ids = set(os.path.join(doc_id, output.id + '.pdf') for output in doc.outputs)
        plot_ids = set(os.path.relpath(file.archive_path, './') for file in archive.files)

        missing_plot_ids = expected_plot_ids.difference(plot_ids)
        extra_plot_ids = plot_ids.difference(expected_plot_ids)

        if missing_plot_ids:
            shutil.rmtree(tempdir)
            raise InvalidOutputsException('Simulator did not produce the following plots:\n  - {}'.format(
                '\n  - '.join(sorted('`' + id + '`' for id in missing_plot_ids))
            ))

        if extra_plot_ids:
            msg = 'Simulator produced extra plots:\n  - {}'.format(
                '\n  - '.join(sorted('`' + id + '`' for id in extra_plot_ids)))
            warnings.warn(msg, InvalidOutputsWarning)

        # check plot data saved
        expected_plot_ids = set()
        for doc_location, doc in synthetic_sed_docs.items():
            doc_id = os.path.relpath(doc_location, './')
            for output in doc.outputs:
                if isinstance(output, (Plot2D, Plot3D)):
                    expected_plot_ids.add(os.path.join(doc_id, output.id))

        try:
            plot_ids = ReportReader().get_ids(outputs_dir)
        except Exception:
            plot_ids = []

        missing_plot_ids = expected_plot_ids.difference(set(plot_ids))

        if missing_plot_ids:
            raise InvalidOutputsException('Simulator did not produce data for the following plots:\n  - {}'.format(
                '\n  - '.join(sorted('`' + id + '`' for id in missing_plot_ids))
            ))

        # remove temporary directory
        shutil.rmtree(tempdir)


class SimulatorProduces2DPlotsTestCase(SimulatorProducesPlotsTestCase):
    """ Test that a simulator produces 2D plots """

    def is_curated_sed_report_suitable_for_building_synthetic_archive(self, specifications, report, sed_doc_location):
        """ Determine if a SED report is suitable for testing

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            report (:obj:`Report`): SED report in curated archive
            sed_doc_location (:obj:`str`): location of the SED document within its parent COMBINE/OMEX archive

        Returns:
            :obj:`bool`: whether the report is suitable for testing
        """
        if not super(SimulatorProduces2DPlotsTestCase, self).is_curated_sed_report_suitable_for_building_synthetic_archive(
                specifications, report, sed_doc_location):
            return False

        sed_doc_id = os.path.relpath(sed_doc_location, '.')
        expected_report = next((expected_report for expected_report in self._published_projects_test_case.expected_reports
                                if expected_report.id == os.path.join(sed_doc_id, report.id)), None)

        if expected_report is None:
            return False

        return len(expected_report.points) == 1

    def build_plots(self, data_generators):
        """ Build plots from the defined data generators

        Args:
            data_generators (:obj:`list` of :obj:`DataGenerator`): data generators

        Returns:
            :obj:`list` of :obj:`Output`: plots
        """
        plots = []
        for i in range(self._num_plots):
            plots.append(Plot2D(id='plot_' + str(i)))

        for i_data_generator, data_generator in enumerate(data_generators):
            plots[i_data_generator % self._num_plots].curves.append(
                Curve(
                    id='curve_' + str(i_data_generator),
                    x_data_generator=data_generator,
                    y_data_generator=data_generator,
                    x_scale=self._axis_scale,
                    y_scale=self._axis_scale,
                ),
            )

        return plots


class SimulatorProduces3DPlotsTestCase(SimulatorProducesPlotsTestCase):
    """ Test that a simulator produces 3D plots """

    def is_curated_sed_report_suitable_for_building_synthetic_archive(self, specifications, report, sed_doc_location):
        """ Determine if a SED report is suitable for testing

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            report (:obj:`Report`): SED report in curated archive
            sed_doc_location (:obj:`str`): location of the SED document within its parent COMBINE/OMEX archive

        Returns:
            :obj:`bool`: whether the report is suitable for testing
        """
        if not super(SimulatorProduces3DPlotsTestCase, self).is_curated_sed_report_suitable_for_building_synthetic_archive(
                specifications, report, sed_doc_location):
            return False

        sed_doc_id = os.path.relpath(sed_doc_location, '.')
        expected_report = next((expected_report for expected_report in self._published_projects_test_case.expected_reports
                                if expected_report.id == os.path.join(sed_doc_id, report.id)), None)
        if expected_report is None:
            return False

        return len(expected_report.points) == 2

    def build_plots(self, data_generators):
        """ Build plots from the defined data generators

        Args:
            data_generators (:obj:`list` of :obj:`DataGenerator`): data generators

        Returns:
            :obj:`list` of :obj:`Output`: plots
        """
        plots = []
        for i in range(self._num_plots):
            plots.append(Plot3D(id='plot_' + str(i)))

        for i_data_generator, data_generator in enumerate(data_generators):
            plots[i_data_generator % self._num_plots].surfaces.append(
                Surface(
                    id='surface_' + str(i_data_generator),
                    x_data_generator=data_generator,
                    y_data_generator=data_generator,
                    z_data_generator=data_generator,
                    x_scale=self._axis_scale,
                    y_scale=self._axis_scale,
                    z_scale=self._axis_scale,
                ),
            )

        return plots


class SimulatorProducesLinear2DPlots(SimulatorProduces2DPlotsTestCase):
    """ Test that a simulator produces linear 2D plots """

    @property
    def _axis_scale(self):
        return AxisScale.linear


class SimulatorProducesLogarithmic2DPlots(SimulatorProduces2DPlotsTestCase):
    """ Test that a simulator produces logarithmic 2D plots """

    @property
    def _axis_scale(self):
        return AxisScale.log


class SimulatorProducesLinear3DPlots(SimulatorProduces3DPlotsTestCase):
    """ Test that a simulator produces linear 3D plots """

    @property
    def _axis_scale(self):
        return AxisScale.linear


class SimulatorProducesLogarithmic3DPlots(SimulatorProduces3DPlotsTestCase):
    """ Test that a simulator produces logarithmic 3D plots """

    @property
    def _axis_scale(self):
        return AxisScale.log


class SimulatorProducesMultiplePlots(SimulatorProduces2DPlotsTestCase):
    """ Test that a simulator produces multiple plots """

    @property
    def _num_plots(self):
        return 2

    @property
    def _axis_scale(self):
        return AxisScale.linear


class SimulatorSupportsDataGeneratorsWithDifferentShapes(UniformTimeCourseTestCase):
    """ Test that a simulator supports data generators with different shapes """
    REPORT_ERROR_AS_SKIP = True

    def modify_simulation(self, simulation):
        """ Modify a simulation

        Args:
            simulation (:obj:`UniformTimeCourseSimulation`): simulation
        """
        pass  # pragma: no cover

    def build_synthetic_archives(self, specifications, curated_archive, curated_archive_dir, curated_sed_docs):
        """ Generate a synthetic archive with a copy of each task and each report

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            curated_archive (:obj:`CombineArchive`): curated COMBINE/OMEX archive
            curated_archive_dir (:obj:`str`): directory with the contents of the curated COMBINE/OMEX archive
            curated_sed_docs (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): map from locations to
                SED documents in curated archive

        Returns:
            :obj:`list` of :obj:`ExpectedResultOfSyntheticArchive`
        """
        expected_results_of_synthetic_archives = super(SimulatorSupportsDataGeneratorsWithDifferentShapes, self).build_synthetic_archives(
            specifications, curated_archive, curated_archive_dir, curated_sed_docs)
        curated_sed_docs = expected_results_of_synthetic_archives[0].sed_documents
        doc = list(curated_sed_docs.values())[0]

        sim = doc.simulations[0]
        sim2 = UniformTimeCourseSimulation(
            id=sim.id + '__copy_2',
            initial_time=sim.initial_time,
            output_start_time=sim.output_start_time,
            output_end_time=sim.output_end_time + (sim.output_end_time - sim.output_start_time),
            number_of_points=sim.number_of_points * 2,
            algorithm=copy.deepcopy(sim.algorithm),
        )
        doc.simulations.append(sim2)

        task = doc.tasks[0]
        task2 = Task(
            id=task.id + '__copy_2',
            model=task.model,
            simulation=sim2,
        )
        doc.tasks.append(task2)

        report = doc.outputs.pop()
        report2 = Report(id=report.id + '__copy_2')
        doc.outputs.append(report2)

        for data_gen in list(doc.data_generators):
            data_gen2 = DataGenerator(
                id=data_gen.id + '__copy_2',
                variables=[
                    Variable(
                        id=data_gen.variables[0].id + '__copy_1',
                        task=task,
                        symbol=data_gen.variables[0].symbol,
                        target=data_gen.variables[0].target,
                        target_namespaces=data_gen.variables[0].target_namespaces,
                    ),
                    Variable(
                        id=data_gen.variables[0].id + '__copy_2',
                        task=task2,
                        symbol=data_gen.variables[0].symbol,
                        target=data_gen.variables[0].target,
                        target_namespaces=data_gen.variables[0].target_namespaces,
                    ),
                ],
                math='{} + {}'.format(data_gen.variables[0].id + '__copy_1', data_gen.variables[0].id + '__copy_2'),
            )
            doc.data_generators.append(data_gen2)

            report2.data_sets.append(DataSet(id=data_gen2.id + '_data_set', label=data_gen2.id + '_data_set', data_generator=data_gen2))

        return expected_results_of_synthetic_archives

    def eval_outputs(self, specifications, synthetic_archive, synthetic_sed_docs, outputs_dir):
        """ Test that the expected outputs were created for the synthetic archive

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            synthetic_archive (:obj:`CombineArchive`): synthetic COMBINE/OMEX archive for testing the simulator
            synthetic_sed_docs (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): map from the location of each SED
                document in the synthetic archive to the document
            outputs_dir (:obj:`str`): directory that contains the outputs produced from the execution of the synthetic archive

        Returns:
            :obj:`bool`: whether there were no warnings about the outputs
        """
        doc_location = list(synthetic_sed_docs.keys())[0]
        doc_id = os.path.relpath(doc_location, './')
        doc = synthetic_sed_docs[doc_location]
        sim1 = doc.simulations[0]
        sim2 = doc.simulations[-1]
        report2 = doc.outputs[-1]

        results2 = ReportReader().run(report2, outputs_dir, os.path.join(doc_id, report2.id))
        for value in results2.values():
            self._eval_data_set(value, sim2.number_of_points + 1, sim1.number_of_points + 1)

    def _eval_data_set(self, value, length, non_nan_points):
        if value.shape[-1] != length:
            raise InvalidOutputsException('Data set does not have the expected shape')

        data_set_slice = tuple([slice(0, dim_len) for dim_len in value.shape[0:-1]] + [slice(0, non_nan_points)])
        if numpy.any(numpy.isnan(value[data_set_slice])):
            raise InvalidOutputsException('Data set has unexpected NaN values')

        data_set_slice = tuple(
            [slice(0, dim_len) for dim_len in value.shape[0:-1]]
            + [slice(non_nan_points, length)]
        )
        if not numpy.all(numpy.isnan(value[data_set_slice])):
            raise InvalidOutputsException('Data set has unexpected non-NaN values')


class SimulatorSupportsDataSetsWithDifferentShapes(UniformTimeCourseTestCase):
    """ Test that a simulator supports data generators with different shapes """
    REPORT_ERROR_AS_SKIP = True

    def modify_simulation(self, simulation):
        """ Modify a simulation

        Args:
            simulation (:obj:`UniformTimeCourseSimulation`): simulation
        """
        pass  # pragma: no cover

    def build_synthetic_archives(self, specifications, curated_archive, curated_archive_dir, curated_sed_docs):
        """ Generate a synthetic archive with a copy of each task and each report

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            curated_archive (:obj:`CombineArchive`): curated COMBINE/OMEX archive
            curated_archive_dir (:obj:`str`): directory with the contents of the curated COMBINE/OMEX archive
            curated_sed_docs (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): map from locations to
                SED documents in curated archive

        Returns:
            :obj:`list` of :obj:`ExpectedResultOfSyntheticArchive`
        """
        expected_results_of_synthetic_archives = super(SimulatorSupportsDataSetsWithDifferentShapes, self).build_synthetic_archives(
            specifications, curated_archive, curated_archive_dir, curated_sed_docs)
        curated_sed_docs = expected_results_of_synthetic_archives[0].sed_documents
        doc = list(curated_sed_docs.values())[0]

        sim = doc.simulations[0]
        sim2 = UniformTimeCourseSimulation(
            id=sim.id + '__copy_2',
            initial_time=sim.initial_time,
            output_start_time=sim.output_start_time,
            output_end_time=sim.output_end_time + (sim.output_end_time - sim.output_start_time),
            number_of_points=sim.number_of_points * 2,
            algorithm=copy.deepcopy(sim.algorithm),
        )
        doc.simulations.append(sim2)

        task = doc.tasks[0]
        task2 = Task(
            id=task.id + '__copy_2',
            model=task.model,
            simulation=sim2,
        )
        doc.tasks.append(task2)

        report = doc.outputs[0]

        for data_gen in list(doc.data_generators):
            data_gen2 = DataGenerator(
                id=data_gen.id + '__copy_2',
                variables=[
                    Variable(
                        id=data_gen.variables[0].id + '__copy_2',
                        task=task2,
                        symbol=data_gen.variables[0].symbol,
                        target=data_gen.variables[0].target,
                        target_namespaces=data_gen.variables[0].target_namespaces,
                    ),
                ],
                math='{}'.format(data_gen.variables[0].id + '__copy_2'),
            )
            doc.data_generators.append(data_gen2)

            report.data_sets.append(DataSet(id=data_gen2.id + '_data_set', label=data_gen2.id + '_data_set', data_generator=data_gen2))

        return expected_results_of_synthetic_archives

    def eval_outputs(self, specifications, synthetic_archive, synthetic_sed_docs, outputs_dir):
        """ Test that the expected outputs were created for the synthetic archive

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            synthetic_archive (:obj:`CombineArchive`): synthetic COMBINE/OMEX archive for testing the simulator
            synthetic_sed_docs (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): map from the location of each SED
                document in the synthetic archive to the document
            outputs_dir (:obj:`str`): directory that contains the outputs produced from the execution of the synthetic archive

        Returns:
            :obj:`bool`: whether there were no warnings about the outputs
        """
        doc_location = list(synthetic_sed_docs.keys())[0]
        doc_id = os.path.relpath(doc_location, './')
        doc = synthetic_sed_docs[doc_location]
        sim1 = doc.simulations[0]
        sim2 = doc.simulations[-1]
        task1 = doc.tasks[0]
        report = doc.outputs[0]

        results = ReportReader().run(report, outputs_dir, os.path.join(doc_id, report.id))

        values1 = {}
        values2 = {}

        for data_set in report.data_sets:
            value = results[data_set.id]

            if data_set.data_generator.variables[0].task == task1:
                expected_length = sim1.number_of_points + 1

                if data_set.data_generator.variables[0].symbol:
                    values1[data_set.data_generator.id] = value

            else:
                expected_length = sim2.number_of_points + 1

                if data_set.data_generator.variables[0].symbol:
                    values2[data_set.data_generator.id.replace('__copy_2', '')] = value[0:sim1.number_of_points + 1]

            self._eval_data_set(data_set.id, value, expected_length)

        for key in values1.keys():
            self._eval_time_data_sets(values1[key], values2[key])

    def _eval_data_set(self, id, value, expected_length):
        if value.shape[-1] != expected_length:
            raise InvalidOutputsException('Data set `{}` does not have the expected shape'.format(id))

        if numpy.any(numpy.isnan(value)):
            raise InvalidOutputsException('Data set `{}` has unexpected NaN values'.format(id))

    def _eval_time_data_sets(self, value1, value2):
        try:
            numpy.testing.assert_allclose(value1, value2)
        except Exception as exception:
            raise InvalidOutputsException('Simulations with the same time courses should produce equivalent time data sets:\n\n  {}'.format(
                str(exception).replace('\n', '\n  ')))


class SimulatorSupportsSubstitutingAlgorithms(SimulatorSupportsModelsSimulationsTasksDataGeneratorsAndReports):
    """ Check that a simulator can substitute algorithms that it doesn't implement with similar
    algorithms when the algorithm substitution policy is less restrictive than
    :obj:`AlgorithmSubstitutionPolicy.SAME_METHOD`. Also check that a simulator ignores unsupported
    algorithm parameters when the algorithm substitution policy is less restrictive than
    :obj:`AlgorithmSubstitutionPolicy.NONE`.
    """
    REPORT_ERROR_AS_SKIP = True

    def is_curated_sed_algorithm_suitable_for_building_synthetic_archive(self, specifications, algorithm):
        """ Determine if a SED algorithm is suitable for testing

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            algorithm (:obj:`Algorithm`): SED algorithm in curated archive

        Returns:
            :obj:`bool`: whether the algorithm is suitable for testing
        """
        kisao = Kisao()
        alg_term = kisao.get_term(algorithm.kisao_id)
        sub_algs = get_substitutable_algorithms_for_policy(
            alg_term,
            substitution_policy=AlgorithmSubstitutionPolicy.SIMILAR_VARIABLES)
        sub_alg_ids = kisao.get_term_ids(sub_algs)

        for alg_specs in specifications['algorithms']:
            if alg_specs['kisaoId']['id'] in sub_alg_ids:
                sub_alg_ids.remove(alg_specs['kisaoId']['id'])

        if sub_alg_ids:
            if 'KISAO_0000019' in sub_alg_ids:
                self._alt_alg = 'KISAO_0000019'
            elif 'KISAO_0000088' in sub_alg_ids:
                self._alt_alg = 'KISAO_0000088'
            else:
                self._alt_alg = sub_alg_ids[0]
            return True
        else:
            return False

    def build_synthetic_archives(self, specifications, curated_archive, curated_archive_dir, curated_sed_docs):
        """ Generate a synthetic archive with master and non-master SED documents

        Args:
            specifications (:obj:`dict`): specifications of the simulator to validate
            curated_archive (:obj:`CombineArchive`): curated COMBINE/OMEX archive
            curated_archive_dir (:obj:`str`): directory with the contents of the curated COMBINE/OMEX archive
            curated_sed_docs (:obj:`dict` of :obj:`str` to :obj:`SedDocument`): map from locations to
                SED documents in curated archive

        Returns:
            :obj:`list` of :obj:`ExpectedResultOfSyntheticArchive`
        """
        super(SimulatorSupportsSubstitutingAlgorithms, self).build_synthetic_archives(
            specifications, curated_archive, curated_archive_dir, curated_sed_docs)

        sed_docs_1 = copy.deepcopy(curated_sed_docs)
        doc = list(sed_docs_1.values())[0]
        doc.simulations[0].algorithm.kisao_id = self._alt_alg
        doc.simulations[0].algorithm.changes = []

        sed_docs_2 = copy.deepcopy(curated_sed_docs)
        doc = list(sed_docs_2.values())[0]
        doc.simulations[0].algorithm.changes.append(AlgorithmParameterChange(
            kisao_id='KISAO_0000428',
            new_value='0.1',
        ))

        return [
            ExpectedResultOfSyntheticArchive(
                curated_archive,
                sed_docs_1,
                False,
                environment={'ALGORITHM_SUBSTITUTION_POLICY': AlgorithmSubstitutionPolicy.NONE.name}),
            ExpectedResultOfSyntheticArchive(
                curated_archive,
                sed_docs_1,
                True,
                environment={'ALGORITHM_SUBSTITUTION_POLICY': AlgorithmSubstitutionPolicy.SIMILAR_VARIABLES.name}),
            ExpectedResultOfSyntheticArchive(
                curated_archive,
                sed_docs_2,
                False,
                environment={'ALGORITHM_SUBSTITUTION_POLICY': AlgorithmSubstitutionPolicy.NONE.name}),
            ExpectedResultOfSyntheticArchive(
                curated_archive,
                sed_docs_2,
                True,
                environment={'ALGORITHM_SUBSTITUTION_POLICY': AlgorithmSubstitutionPolicy.SIMILAR_VARIABLES.name}),
        ]



""" Utility methods

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-01-02
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

__all__ = ['are_array_shapes_equivalent', 'reduce_array_shape']


def are_array_shapes_equivalent(shape_1, shape_2, same_dims=False):
    """ Determine if two NumPy arrays have equivalent shape

    * Trim trailing ones
    * Check that remaining non-zero dimension sizes are equal
    * If :obj:`same_dims` is :obj:`True`, all check that the length of the shapes are equal

    Args:
        shape_1 (:obj:`list` of :obj:`int`): shape of first array
        shape_2 (:obj:`list` of :obj:`int`): shape of second array
        same_dims (:obj:`bool`, optional): if :obj:`True`, check that the dimensions
            are also the same

    Returns:
        :obj:`bool`: :obj:`True` if the shapes are equivalent
    """
    reduced_shape_1 = reduce_array_shape(shape_1)
    reduced_shape_2 = reduce_array_shape(shape_2)

    if reduced_shape_1 != reduced_shape_2:
        return False

    if same_dims and len(shape_1) != len(shape_2):
        return False

    return True


def reduce_array_shape(shape):
    """ Reduce the shape of a NumPy array

    * Trim trailing ones
    * Check that remaining non-zero dimension sizes are equal

    Args:
        shape (:obj:`list` of :obj:`int`): shape of array

    Returns:
        :obj:`:obj:`list` of :obj:`int`: reduced shape
    """
    shape = list(shape)

    while shape and shape[-1] == 1:
        shape.pop()

    return shape
